# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a list."""
from collections.abc import MutableSequence

from .base import to_aiida_type
from .data import Data

__all__ = ('List',)


class List(Data, MutableSequence):
    """`Data` sub class to represent a list."""

    _LIST_KEY = 'list'

    def __init__(self, value=None, **kwargs):
        """Initialise a ``List`` node instance.

        :param value: list to initialise the ``List`` node from
        """
        data = value or kwargs.pop('list', [])
        super().__init__(**kwargs)
        self.set_list(data)

    def __getitem__(self, item):
        return self.get_list()[item]

    def __setitem__(self, key, value):
        data = self.get_list()
        data[key] = value
        if not self._using_list_reference():
            self.set_list(data)

    def __delitem__(self, key):
        data = self.get_list()
        del data[key]
        if not self._using_list_reference():
            self.set_list(data)

    def __len__(self):
        return len(self.get_list())

    def __str__(self):
        return f'{super().__str__()} value: {self.get_list()}'

    def __eq__(self, other):
        if isinstance(other, List):
            return self.get_list() == other.get_list()
        return self.get_list() == other

    def append(self, value):
        data = self.get_list()
        data.append(value)
        if not self._using_list_reference():
            self.set_list(data)

    def extend(self, value):  # pylint: disable=arguments-renamed
        data = self.get_list()
        data.extend(value)
        if not self._using_list_reference():
            self.set_list(data)

    def insert(self, i, value):  # pylint: disable=arguments-renamed
        data = self.get_list()
        data.insert(i, value)
        if not self._using_list_reference():
            self.set_list(data)

    def remove(self, value):
        data = self.get_list()
        item = data.remove(value)
        if not self._using_list_reference():
            self.set_list(data)
        return item

    def pop(self, **kwargs):  # pylint: disable=arguments-differ
        """Remove and return item at index (default last)."""
        data = self.get_list()
        item = data.pop(**kwargs)
        if not self._using_list_reference():
            self.set_list(data)
        return item

    def index(self, value):  # pylint: disable=arguments-differ
        """Return first index of value.."""
        return self.get_list().index(value)

    def count(self, value):
        """Return number of occurrences of value."""
        return self.get_list().count(value)

    def sort(self, key=None, reverse=False):
        data = self.get_list()
        data.sort(key=key, reverse=reverse)
        if not self._using_list_reference():
            self.set_list(data)

    def reverse(self):
        data = self.get_list()
        data.reverse()
        if not self._using_list_reference():
            self.set_list(data)

    def get_list(self):
        """Return the contents of this node.

        :return: a list
        """
        try:
            return self.get_attribute(self._LIST_KEY)
        except AttributeError:
            self.set_list([])
            return self.get_attribute(self._LIST_KEY)

    def set_list(self, data):
        """Set the contents of this node.

        :param data: the list to set
        """
        if not isinstance(data, list):
            raise TypeError('Must supply list type')
        self.set_attribute(self._LIST_KEY, data.copy())

    def _using_list_reference(self):
        """
        This function tells the class if we are using a list reference.  This
        means that calls to self.get_list return a reference rather than a copy
        of the underlying list and therefore self.set_list need not be called.
        This knwoledge is essential to make sure this class is performant.

        Currently the implementation assumes that if the node needs to be
        stored then it is using the attributes cache which is a reference.

        :return: True if using self.get_list returns a reference to the
            underlying sequence.  False otherwise.
        :rtype: bool
        """
        return not self.is_stored


@to_aiida_type.register(list)
def _(value):
    return List(list=value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base settings required for the configuration of an AiiDA instance."""
import os
import pathlib
import typing
import warnings

USE_TZ = True

DEFAULT_UMASK = 0o0077
DEFAULT_AIIDA_PATH_VARIABLE = 'AIIDA_PATH'
DEFAULT_AIIDA_PATH = '~'
DEFAULT_AIIDA_USER = 'aiida@localhost'
DEFAULT_CONFIG_DIR_NAME = '.aiida'
DEFAULT_CONFIG_FILE_NAME = 'config.json'
DEFAULT_CONFIG_INDENT_SIZE = 4
DEFAULT_DAEMON_DIR_NAME = 'daemon'
DEFAULT_DAEMON_LOG_DIR_NAME = 'log'
DEFAULT_ACCESS_CONTROL_DIR_NAME = 'access'

AIIDA_CONFIG_FOLDER: typing.Optional[pathlib.Path] = None
DAEMON_DIR: typing.Optional[pathlib.Path] = None
DAEMON_LOG_DIR: typing.Optional[pathlib.Path] = None
ACCESS_CONTROL_DIR: typing.Optional[pathlib.Path] = None


def create_instance_directories():
    """Create the base directories required for a new AiiDA instance.

    This will create the base AiiDA directory defined by the AIIDA_CONFIG_FOLDER variable, unless it already exists.
    Subsequently, it will create the daemon directory within it and the daemon log directory.
    """
    from aiida.common import ConfigurationError

    directory_base = pathlib.Path(AIIDA_CONFIG_FOLDER).expanduser()
    directory_daemon = directory_base / DAEMON_DIR
    directory_daemon_log = directory_base / DAEMON_LOG_DIR
    directory_access = directory_base / ACCESS_CONTROL_DIR

    list_of_paths = [
        directory_base,
        directory_daemon,
        directory_daemon_log,
        directory_access,
    ]

    umask = os.umask(DEFAULT_UMASK)

    try:
        for path in list_of_paths:

            if path is directory_base and not path.exists():
                warnings.warn(f'Creating AiiDA configuration folder `{path}`.')

            try:
                path.mkdir(parents=True, exist_ok=True)
            except OSError as exc:
                raise ConfigurationError(f'could not create the `{path}` configuration directory: {exc}') from exc
    finally:
        os.umask(umask)


def set_configuration_directory():
    """Determine the location of the configuration directory and set the related global variables.

    The location of the configuration folder will be determined and optionally created following these heuristics:

        * If the `AIIDA_PATH` variable is set, all the paths will be checked to see if they contain a configuration
          folder. The first one to be encountered will be set as `AIIDA_CONFIG_FOLDER`. If none of them contain one,
          a configuration folder will be created in the last path considered.
        * If the `AIIDA_PATH` variable is not set the `DEFAULT_AIIDA_PATH` value will be used as base path and if it
          does not yet contain a configuration folder, one will be created.

    In principle then, a configuration folder should always be found or automatically created.
    """
    # pylint: disable = global-statement
    global AIIDA_CONFIG_FOLDER
    global DAEMON_DIR
    global DAEMON_LOG_DIR
    global ACCESS_CONTROL_DIR

    environment_variable = os.environ.get(DEFAULT_AIIDA_PATH_VARIABLE, None)

    if environment_variable:

        # Loop over all the paths in the `AIIDA_PATH` variable to see if any of them contain a configuration folder
        for base_dir_path in [path for path in environment_variable.split(':') if path]:

            AIIDA_CONFIG_FOLDER = pathlib.Path(base_dir_path).expanduser()

            # Only add the base config directory name to the base path if it does not already do so
            # Someone might already include it in the environment variable. e.g.: AIIDA_PATH=/home/some/path/.aiida
            if AIIDA_CONFIG_FOLDER.name != DEFAULT_CONFIG_DIR_NAME:
                AIIDA_CONFIG_FOLDER = AIIDA_CONFIG_FOLDER / DEFAULT_CONFIG_DIR_NAME

            # If the directory exists, we leave it set and break the loop
            if AIIDA_CONFIG_FOLDER.is_dir():
                break

    else:
        # The `AIIDA_PATH` variable is not set, so default to the default path and try to create it if it does not exist
        AIIDA_CONFIG_FOLDER = pathlib.Path(DEFAULT_AIIDA_PATH).expanduser() / DEFAULT_CONFIG_DIR_NAME

    DAEMON_DIR = AIIDA_CONFIG_FOLDER / DEFAULT_DAEMON_DIR_NAME
    DAEMON_LOG_DIR = DAEMON_DIR / DEFAULT_DAEMON_LOG_DIR_NAME
    ACCESS_CONTROL_DIR = AIIDA_CONFIG_FOLDER / DEFAULT_ACCESS_CONTROL_DIR_NAME

    create_instance_directories()


# Initialize the configuration directory settings
set_configuration_directory()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unittests for TestManager"""
import sys
import unittest
import warnings

import pytest

from aiida.manage.tests import TestManager, get_test_backend_name


class TestManagerTestCase(unittest.TestCase):
    """Test the TestManager class"""

    def setUp(self):
        if sys.version_info[0] >= 3:
            # tell unittest not to warn about running processes
            warnings.simplefilter('ignore', ResourceWarning)  # pylint: disable=no-member,undefined-variable

        self.backend = get_test_backend_name()
        self.test_manager = TestManager()

    def tearDown(self):
        self.test_manager.destroy_all()

    @pytest.mark.filterwarnings('ignore:Creating AiiDA configuration folder')
    def test_pgtest_argument(self):
        """
        Create a temporary profile, passing the pgtest argument.
        """
        from pgtest.pgtest import which

        # this should fail
        pgtest = {'pg_ctl': 'notapath'}
        with self.assertRaises(AssertionError):
            self.test_manager.use_temporary_profile(backend=self.backend, pgtest=pgtest)

        # pg_ctl is what PGTest also looks for (although it might be more clever)
        pgtest = {'pg_ctl': which('pg_ctl')}
        self.test_manager.use_temporary_profile(backend=self.backend, pgtest=pgtest)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Main module to expose all orm classes and methods"""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .authinfos import *
from .comments import *
from .computers import *
from .entities import *
from .groups import *
from .logs import *
from .nodes import *
from .querybuilder import *
from .users import *
from .utils import *

__all__ = (
    'ASCENDING',
    'AbstractNodeMeta',
    'ArrayData',
    'AttributeManager',
    'AuthInfo',
    'AutoGroup',
    'BandsData',
    'BaseType',
    'Bool',
    'CalcFunctionNode',
    'CalcJobNode',
    'CalcJobResultManager',
    'CalculationEntityLoader',
    'CalculationNode',
    'CifData',
    'Code',
    'CodeEntityLoader',
    'Collection',
    'Comment',
    'Computer',
    'ComputerEntityLoader',
    'DESCENDING',
    'Data',
    'Dict',
    'Entity',
    'EntityAttributesMixin',
    'EntityExtrasMixin',
    'EntityTypes',
    'EnumData',
    'Float',
    'FolderData',
    'Group',
    'GroupEntityLoader',
    'ImportGroup',
    'Int',
    'JsonableData',
    'Kind',
    'KpointsData',
    'LinkManager',
    'LinkPair',
    'LinkTriple',
    'List',
    'Log',
    'Node',
    'NodeEntityLoader',
    'NodeLinksManager',
    'NodeRepositoryMixin',
    'NumericType',
    'OrbitalData',
    'OrderSpecifier',
    'OrmEntityLoader',
    'ProcessNode',
    'ProjectionData',
    'QueryBuilder',
    'RemoteData',
    'RemoteStashData',
    'RemoteStashFolderData',
    'SinglefileData',
    'Site',
    'Str',
    'StructureData',
    'TrajectoryData',
    'UpfData',
    'UpfFamily',
    'User',
    'WorkChainNode',
    'WorkFunctionNode',
    'WorkflowNode',
    'XyData',
    'cif_from_ase',
    'find_bandgap',
    'get_loader',
    'get_query_type_from_type_string',
    'get_type_string_from_class',
    'has_pycifrw',
    'load_code',
    'load_computer',
    'load_entity',
    'load_group',
    'load_node',
    'load_node_class',
    'pycifrw_from_cif',
    'to_aiida_type',
    'validate_link',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This file contains the main routines to submit, check and retrieve calculation
results. These are general and contain only the main logic; where appropriate,
the routines make reference to the suitable plugins for all
plugin-specific operations.
"""
from collections.abc import Mapping
from logging import LoggerAdapter
import os
import pathlib
import shutil
from tempfile import NamedTemporaryFile
from typing import Any, List
from typing import Mapping as MappingType
from typing import Optional, Tuple, Union

from aiida.common import AIIDA_LOGGER, exceptions
from aiida.common.datastructures import CalcInfo
from aiida.common.folders import SandboxFolder
from aiida.common.links import LinkType
from aiida.orm import CalcJobNode, Code, FolderData, Node, RemoteData, load_node
from aiida.orm.utils.log import get_dblogger_extra
from aiida.repository.common import FileType
from aiida.schedulers.datastructures import JobState
from aiida.transports import Transport

REMOTE_WORK_DIRECTORY_LOST_FOUND = 'lost+found'

EXEC_LOGGER = AIIDA_LOGGER.getChild('execmanager')


def _find_data_node(inputs: MappingType[str, Any], uuid: str) -> Optional[Node]:
    """Find and return the node with the given UUID from a nested mapping of input nodes.

    :param inputs: (nested) mapping of nodes
    :param uuid: UUID of the node to find
    :return: instance of `Node` or `None` if not found
    """
    data_node = None

    for input_node in inputs.values():
        if isinstance(input_node, Mapping):
            data_node = _find_data_node(input_node, uuid)
        elif isinstance(input_node, Node) and input_node.uuid == uuid:
            data_node = input_node
        if data_node is not None:
            break

    return data_node


def upload_calculation(
    node: CalcJobNode,
    transport: Transport,
    calc_info: CalcInfo,
    folder: SandboxFolder,
    inputs: Optional[MappingType[str, Any]] = None,
    dry_run: bool = False
) -> None:
    """Upload a `CalcJob` instance

    :param node: the `CalcJobNode`.
    :param transport: an already opened transport to use to submit the calculation.
    :param calc_info: the calculation info datastructure returned by `CalcJob.presubmit`
    :param folder: temporary local file system folder containing the inputs written by `CalcJob.prepare_for_submission`
    """
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements

    # If the calculation already has a `remote_folder`, simply return. The upload was apparently already completed
    # before, which can happen if the daemon is restarted and it shuts down after uploading but before getting the
    # chance to perform the state transition. Upon reloading this calculation, it will re-attempt the upload.
    link_label = 'remote_folder'
    if node.get_outgoing(RemoteData, link_label_filter=link_label).first():
        EXEC_LOGGER.warning(f'CalcJobNode<{node.pk}> already has a `{link_label}` output: skipping upload')
        return calc_info

    computer = node.computer

    codes_info = calc_info.codes_info
    input_codes = [load_node(_.code_uuid, sub_classes=(Code,)) for _ in codes_info]

    logger_extra = get_dblogger_extra(node)
    transport.set_logger_extra(logger_extra)
    logger = LoggerAdapter(logger=EXEC_LOGGER, extra=logger_extra)

    if not dry_run and node.has_cached_links():
        raise ValueError(
            'Cannot submit calculation {} because it has cached input links! If you just want to test the '
            'submission, set `metadata.dry_run` to True in the inputs.'.format(node.pk)
        )

    # If we are performing a dry-run, the working directory should actually be a local folder that should already exist
    if dry_run:
        workdir = transport.getcwd()
    else:
        remote_user = transport.whoami()
        remote_working_directory = computer.get_workdir().format(username=remote_user)
        if not remote_working_directory.strip():
            raise exceptions.ConfigurationError(
                "[submission of calculation {}] No remote_working_directory configured for computer '{}'".format(
                    node.pk, computer.label
                )
            )

        # If it already exists, no exception is raised
        try:
            transport.chdir(remote_working_directory)
        except IOError:
            logger.debug(
                '[submission of calculation {}] Unable to chdir in {}, trying to create it'.format(
                    node.pk, remote_working_directory
                )
            )
            try:
                transport.makedirs(remote_working_directory)
                transport.chdir(remote_working_directory)
            except EnvironmentError as exc:
                raise exceptions.ConfigurationError(
                    '[submission of calculation {}] '
                    'Unable to create the remote directory {} on '
                    "computer '{}': {}".format(node.pk, remote_working_directory, computer.label, exc)
                )
        # Store remotely with sharding (here is where we choose
        # the folder structure of remote jobs; then I store this
        # in the calculation properties using _set_remote_dir
        # and I do not have to know the logic, but I just need to
        # read the absolute path from the calculation properties.
        transport.mkdir(calc_info.uuid[:2], ignore_existing=True)
        transport.chdir(calc_info.uuid[:2])
        transport.mkdir(calc_info.uuid[2:4], ignore_existing=True)
        transport.chdir(calc_info.uuid[2:4])

        try:
            # The final directory may already exist, most likely because this function was already executed once, but
            # failed and as a result was rescheduled by the eninge. In this case it would be fine to delete the folder
            # and create it from scratch, except that we cannot be sure that this the actual case. Therefore, to err on
            # the safe side, we move the folder to the lost+found directory before recreating the folder from scratch
            transport.mkdir(calc_info.uuid[4:])
        except OSError:
            # Move the existing directory to lost+found, log a warning and create a clean directory anyway
            path_existing = os.path.join(transport.getcwd(), calc_info.uuid[4:])
            path_lost_found = os.path.join(remote_working_directory, REMOTE_WORK_DIRECTORY_LOST_FOUND)
            path_target = os.path.join(path_lost_found, calc_info.uuid)
            logger.warning(
                f'tried to create path {path_existing} but it already exists, moving the entire folder to {path_target}'
            )

            # Make sure the lost+found directory exists, then copy the existing folder there and delete the original
            transport.mkdir(path_lost_found, ignore_existing=True)
            transport.copytree(path_existing, path_target)
            transport.rmtree(path_existing)

            # Now we can create a clean folder for this calculation
            transport.mkdir(calc_info.uuid[4:])
        finally:
            transport.chdir(calc_info.uuid[4:])

        # I store the workdir of the calculation for later file retrieval
        workdir = transport.getcwd()
        node.set_remote_workdir(workdir)

    # I first create the code files, so that the code can put
    # default files to be overwritten by the plugin itself.
    # Still, beware! The code file itself could be overwritten...
    # But I checked for this earlier.
    for code in input_codes:
        if code.is_local():
            # Note: this will possibly overwrite files
            for filename in code.list_object_names():
                # Note, once #2579 is implemented, use the `node.open` method instead of the named temporary file in
                # combination with the new `Transport.put_object_from_filelike`
                # Since the content of the node could potentially be binary, we read the raw bytes and pass them on
                with NamedTemporaryFile(mode='wb+') as handle:
                    handle.write(code.get_object_content(filename, mode='rb'))
                    handle.flush()
                    transport.put(handle.name, filename)
            transport.chmod(code.get_local_executable(), 0o755)  # rwxr-xr-x

    # local_copy_list is a list of tuples, each with (uuid, dest_path, rel_path)
    # NOTE: validation of these lists are done inside calculation.presubmit()
    local_copy_list = calc_info.local_copy_list or []
    remote_copy_list = calc_info.remote_copy_list or []
    remote_symlink_list = calc_info.remote_symlink_list or []
    provenance_exclude_list = calc_info.provenance_exclude_list or []

    for uuid, filename, target in local_copy_list:
        logger.debug(f'[submission of calculation {node.uuid}] copying local file/folder to {target}')

        try:
            data_node = load_node(uuid=uuid)
        except exceptions.NotExistent:
            data_node = _find_data_node(inputs, uuid) if inputs else None

        if data_node is None:
            logger.warning(f'failed to load Node<{uuid}> specified in the `local_copy_list`')
        else:

            # If no explicit source filename is defined, we assume the top-level directory
            filename_source = filename or '.'
            filename_target = target or ''

            # Make the target filepath absolute and create any intermediate directories if they don't yet exist
            filepath_target = pathlib.Path(folder.abspath) / filename_target
            filepath_target.parent.mkdir(parents=True, exist_ok=True)

            if data_node.get_object(filename_source).file_type == FileType.DIRECTORY:
                # If the source object is a directory, we copy its entire contents
                data_node.copy_tree(filepath_target, filename_source)
                provenance_exclude_list.extend(data_node.list_object_names(filename_source))
            else:
                # Otherwise, simply copy the file
                with folder.open(target, 'wb') as handle:
                    with data_node.open(filename, 'rb') as source:
                        shutil.copyfileobj(source, handle)

                provenance_exclude_list.append(target)

    # In a dry_run, the working directory is the raw input folder, which will already contain these resources
    if not dry_run:
        for filename in folder.get_content_list():
            logger.debug(f'[submission of calculation {node.pk}] copying file/folder {filename}...')
            transport.put(folder.get_abs_path(filename), filename)

        for (remote_computer_uuid, remote_abs_path, dest_rel_path) in remote_copy_list:
            if remote_computer_uuid == computer.uuid:
                logger.debug(
                    '[submission of calculation {}] copying {} remotely, directly on the machine {}'.format(
                        node.pk, dest_rel_path, computer.label
                    )
                )
                try:
                    transport.copy(remote_abs_path, dest_rel_path)
                except (IOError, OSError):
                    logger.warning(
                        '[submission of calculation {}] Unable to copy remote resource from {} to {}! '
                        'Stopping.'.format(node.pk, remote_abs_path, dest_rel_path)
                    )
                    raise
            else:
                raise NotImplementedError(
                    '[submission of calculation {}] Remote copy between two different machines is '
                    'not implemented yet'.format(node.pk)
                )

        for (remote_computer_uuid, remote_abs_path, dest_rel_path) in remote_symlink_list:
            if remote_computer_uuid == computer.uuid:
                logger.debug(
                    '[submission of calculation {}] copying {} remotely, directly on the machine {}'.format(
                        node.pk, dest_rel_path, computer.label
                    )
                )
                try:
                    transport.symlink(remote_abs_path, dest_rel_path)
                except (IOError, OSError):
                    logger.warning(
                        '[submission of calculation {}] Unable to create remote symlink from {} to {}! '
                        'Stopping.'.format(node.pk, remote_abs_path, dest_rel_path)
                    )
                    raise
            else:
                raise IOError(
                    f'It is not possible to create a symlink between two different machines for calculation {node.pk}'
                )
    else:

        if remote_copy_list:
            filepath = os.path.join(workdir, '_aiida_remote_copy_list.txt')
            with open(filepath, 'w', encoding='utf-8') as handle:  # type: ignore[assignment]
                for remote_computer_uuid, remote_abs_path, dest_rel_path in remote_copy_list:
                    handle.write(
                        'would have copied {} to {} in working directory on remote {}'.format(
                            remote_abs_path, dest_rel_path, computer.label
                        )
                    )

        if remote_symlink_list:
            filepath = os.path.join(workdir, '_aiida_remote_symlink_list.txt')
            with open(filepath, 'w', encoding='utf-8') as handle:  # type: ignore[assignment]
                for remote_computer_uuid, remote_abs_path, dest_rel_path in remote_symlink_list:
                    handle.write(
                        'would have created symlinks from {} to {} in working directory on remote {}'.format(
                            remote_abs_path, dest_rel_path, computer.label
                        )
                    )

    # Loop recursively over content of the sandbox folder copying all that are not in `provenance_exclude_list`. Note
    # that directories are not created explicitly. The `node.put_object_from_filelike` call will create intermediate
    # directories for nested files automatically when needed. This means though that empty folders in the sandbox or
    # folders that would be empty when considering the `provenance_exclude_list` will *not* be copied to the repo. The
    # advantage of this explicit copying instead of deleting the files from `provenance_exclude_list` from the sandbox
    # first before moving the entire remaining content to the node's repository, is that in this way we are guaranteed
    # not to accidentally move files to the repository that should not go there at all cost. Note that all entries in
    # the provenance exclude list are normalized first, just as the paths that are in the sandbox folder, otherwise the
    # direct equality test may fail, e.g.: './path/file.txt' != 'path/file.txt' even though they reference the same file
    provenance_exclude_list = [os.path.normpath(entry) for entry in provenance_exclude_list]

    for root, _, filenames in os.walk(folder.abspath):
        for filename in filenames:
            filepath = os.path.join(root, filename)
            relpath = os.path.normpath(os.path.relpath(filepath, folder.abspath))
            dirname = os.path.dirname(relpath)

            # Construct a list of all (partial) filepaths
            # For example, if `relpath == 'some/sub/directory/file.txt'` then the list of relative directory paths is
            # ['some', 'some/sub', 'some/sub/directory']
            # This is necessary, because if any of these paths is in the `provenance_exclude_list` the file should not
            # be copied over.
            components = dirname.split(os.sep)
            dirnames = [os.path.join(*components[:i]) for i in range(1, len(components) + 1)]
            if relpath not in provenance_exclude_list and all(
                dirname not in provenance_exclude_list for dirname in dirnames
            ):
                with open(filepath, 'rb') as handle:  # type: ignore[assignment]
                    node._repository.put_object_from_filelike(handle, relpath)  # pylint: disable=protected-access

    # Since the node is already stored, we cannot use the normal repository interface since it will raise a
    # `ModificationNotAllowed` error. To bypass it, we go straight to the underlying repository instance to store the
    # files, however, this means we have to manually update the node's repository metadata.
    node._update_repository_metadata()  # pylint: disable=protected-access

    if not dry_run:
        # Make sure that attaching the `remote_folder` with a link is the last thing we do. This gives the biggest
        # chance of making this method idempotent. That is to say, if a runner gets interrupted during this action, it
        # will simply retry the upload, unless we got here and managed to link it up, in which case we move to the next
        # task. Because in that case, the check for the existence of this link at the top of this function will exit
        # early from this command.
        remotedata = RemoteData(computer=computer, remote_path=workdir)
        remotedata.add_incoming(node, link_type=LinkType.CREATE, link_label='remote_folder')
        remotedata.store()


def submit_calculation(calculation: CalcJobNode, transport: Transport) -> str:
    """Submit a previously uploaded `CalcJob` to the scheduler.

    :param calculation: the instance of CalcJobNode to submit.
    :param transport: an already opened transport to use to submit the calculation.
    :return: the job id as returned by the scheduler `submit_from_script` call
    """
    job_id = calculation.get_job_id()

    # If the `job_id` attribute is already set, that means this function was already executed once and the scheduler
    # submit command was successful as the job id it returned was set on the node. This scenario can happen when the
    # daemon runner gets shutdown right after accomplishing the submission task, but before it gets the chance to
    # finalize the state transition of the `CalcJob` to the `UPDATE` transport task. Since the job is already submitted
    # we do not want to submit it a second time, so we simply return the existing job id here.
    if job_id is not None:
        return job_id

    scheduler = calculation.computer.get_scheduler()
    scheduler.set_transport(transport)

    submit_script_filename = calculation.get_option('submit_script_filename')
    workdir = calculation.get_remote_workdir()
    job_id = scheduler.submit_from_script(workdir, submit_script_filename)
    calculation.set_job_id(job_id)

    return job_id


def stash_calculation(calculation: CalcJobNode, transport: Transport) -> None:
    """Stash files from the working directory of a completed calculation to a permanent remote folder.

    After a calculation has been completed, optionally stash files from the work directory to a storage location on the
    same remote machine. This is useful if one wants to keep certain files from a completed calculation to be removed
    from the scratch directory, because they are necessary for restarts, but that are too heavy to retrieve.
    Instructions of which files to copy where are retrieved from the `stash.source_list` option.

    :param calculation: the calculation job node.
    :param transport: an already opened transport.
    """
    from aiida.common.datastructures import StashMode
    from aiida.orm import RemoteStashFolderData

    logger_extra = get_dblogger_extra(calculation)

    stash_options = calculation.get_option('stash')
    stash_mode = stash_options.get('mode', StashMode.COPY.value)
    source_list = stash_options.get('source_list', [])

    if not source_list:
        return

    if stash_mode != StashMode.COPY.value:
        EXEC_LOGGER.warning(f'stashing mode {stash_mode} is not implemented yet.')
        return

    cls = RemoteStashFolderData

    EXEC_LOGGER.debug(f'stashing files for calculation<{calculation.pk}>: {source_list}', extra=logger_extra)

    uuid = calculation.uuid
    target_basepath = os.path.join(stash_options['target_base'], uuid[:2], uuid[2:4], uuid[4:])

    for source_filename in source_list:

        source_filepath = os.path.join(calculation.get_remote_workdir(), source_filename)
        target_filepath = os.path.join(target_basepath, source_filename)

        # If the source file is in a (nested) directory, create those directories first in the target directory
        target_dirname = os.path.dirname(target_filepath)
        transport.makedirs(target_dirname, ignore_existing=True)

        try:
            transport.copy(source_filepath, target_filepath)
        except (IOError, ValueError) as exception:
            EXEC_LOGGER.warning(f'failed to stash {source_filepath} to {target_filepath}: {exception}')
        else:
            EXEC_LOGGER.debug(f'stashed {source_filepath} to {target_filepath}')

    remote_stash = cls(
        computer=calculation.computer,
        target_basepath=target_basepath,
        stash_mode=StashMode(stash_mode),
        source_list=source_list,
    ).store()
    remote_stash.add_incoming(calculation, link_type=LinkType.CREATE, link_label='remote_stash')


def retrieve_calculation(calculation: CalcJobNode, transport: Transport, retrieved_temporary_folder: str) -> None:
    """Retrieve all the files of a completed job calculation using the given transport.

    If the job defined anything in the `retrieve_temporary_list`, those entries will be stored in the
    `retrieved_temporary_folder`. The caller is responsible for creating and destroying this folder.

    :param calculation: the instance of CalcJobNode to update.
    :param transport: an already opened transport to use for the retrieval.
    :param retrieved_temporary_folder: the absolute path to a directory in which to store the files
        listed, if any, in the `retrieved_temporary_folder` of the jobs CalcInfo
    """
    logger_extra = get_dblogger_extra(calculation)
    workdir = calculation.get_remote_workdir()

    EXEC_LOGGER.debug(f'Retrieving calc {calculation.pk}', extra=logger_extra)
    EXEC_LOGGER.debug(f'[retrieval of calc {calculation.pk}] chdir {workdir}', extra=logger_extra)

    # If the calculation already has a `retrieved` folder, simply return. The retrieval was apparently already completed
    # before, which can happen if the daemon is restarted and it shuts down after retrieving but before getting the
    # chance to perform the state transition. Upon reloading this calculation, it will re-attempt the retrieval.
    link_label = calculation.link_label_retrieved
    if calculation.get_outgoing(FolderData, link_label_filter=link_label).first():
        EXEC_LOGGER.warning(
            f'CalcJobNode<{calculation.pk}> already has a `{link_label}` output folder: skipping retrieval'
        )
        return

    # Create the FolderData node into which to store the files that are to be retrieved
    retrieved_files = FolderData()

    with transport:
        transport.chdir(workdir)

        # First, retrieve the files of folderdata
        retrieve_list = calculation.get_retrieve_list()
        retrieve_temporary_list = calculation.get_retrieve_temporary_list()

        with SandboxFolder() as folder:
            retrieve_files_from_list(calculation, transport, folder.abspath, retrieve_list)
            # Here I retrieved everything; now I store them inside the calculation
            retrieved_files.put_object_from_tree(folder.abspath)

        # Retrieve the temporary files in the retrieved_temporary_folder if any files were
        # specified in the 'retrieve_temporary_list' key
        if retrieve_temporary_list:
            retrieve_files_from_list(calculation, transport, retrieved_temporary_folder, retrieve_temporary_list)

            # Log the files that were retrieved in the temporary folder
            for filename in os.listdir(retrieved_temporary_folder):
                EXEC_LOGGER.debug(
                    f"[retrieval of calc {calculation.pk}] Retrieved temporary file or folder '{filename}'",
                    extra=logger_extra
                )

        # Store everything
        EXEC_LOGGER.debug(
            f'[retrieval of calc {calculation.pk}] Storing retrieved_files={retrieved_files.pk}', extra=logger_extra
        )
        retrieved_files.store()

    # Make sure that attaching the `retrieved` folder with a link is the last thing we do. This gives the biggest chance
    # of making this method idempotent. That is to say, if a runner gets interrupted during this action, it will simply
    # retry the retrieval, unless we got here and managed to link it up, in which case we move to the next task.
    retrieved_files.add_incoming(calculation, link_type=LinkType.CREATE, link_label=calculation.link_label_retrieved)


def kill_calculation(calculation: CalcJobNode, transport: Transport) -> None:
    """
    Kill the calculation through the scheduler

    :param calculation: the instance of CalcJobNode to kill.
    :param transport: an already opened transport to use to address the scheduler
    """
    job_id = calculation.get_job_id()

    if job_id is None:
        # the calculation has not yet been submitted to the scheduler
        return

    # Get the scheduler plugin class and initialize it with the correct transport
    scheduler = calculation.computer.get_scheduler()
    scheduler.set_transport(transport)

    # Call the proper kill method for the job ID of this calculation
    result = scheduler.kill(job_id)

    if result is not True:

        # Failed to kill because the job might have already been completed
        running_jobs = scheduler.get_jobs(jobs=[job_id], as_dict=True)
        job = running_jobs.get(job_id, None)

        # If the job is returned it is still running and the kill really failed, so we raise
        if job is not None and job.job_state != JobState.DONE:
            raise exceptions.RemoteOperationError(f'scheduler.kill({job_id}) was unsuccessful')
        else:
            EXEC_LOGGER.warning(
                'scheduler.kill() failed but job<{%s}> no longer seems to be running regardless', job_id
            )


def retrieve_files_from_list(
    calculation: CalcJobNode, transport: Transport, folder: str, retrieve_list: List[Union[str, Tuple[str, str, int],
                                                                                           list]]
) -> None:
    """
    Retrieve all the files in the retrieve_list from the remote into the
    local folder instance through the transport. The entries in the retrieve_list
    can be of two types:

        * a string
        * a list

    If it is a string, it represents the remote absolute filepath of the file.
    If the item is a list, the elements will correspond to the following:

        * remotepath
        * localpath
        * depth

    If the remotepath contains file patterns with wildcards, the localpath will be
    treated as the work directory of the folder and the depth integer determines
    upto what level of the original remotepath nesting the files will be copied.

    :param transport: the Transport instance.
    :param folder: an absolute path to a folder that contains the files to copy.
    :param retrieve_list: the list of files to retrieve.
    """
    for item in retrieve_list:
        if isinstance(item, (list, tuple)):
            tmp_rname, tmp_lname, depth = item
            # if there are more than one file I do something differently
            if transport.has_magic(tmp_rname):
                remote_names = transport.glob(tmp_rname)
                local_names = []
                for rem in remote_names:
                    to_append = rem.split(os.path.sep)[-depth:] if depth > 0 else []
                    local_names.append(os.path.sep.join([tmp_lname] + to_append))
            else:
                remote_names = [tmp_rname]
                to_append = tmp_rname.split(os.path.sep)[-depth:] if depth > 0 else []
                local_names = [os.path.sep.join([tmp_lname] + to_append)]
            if depth > 1:  # create directories in the folder, if needed
                for this_local_file in local_names:
                    new_folder = os.path.join(folder, os.path.split(this_local_file)[0])
                    if not os.path.exists(new_folder):
                        os.makedirs(new_folder)
        else:  # it is a string
            if transport.has_magic(item):
                remote_names = transport.glob(item)
                local_names = [os.path.split(rem)[1] for rem in remote_names]
            else:
                remote_names = [item]
                local_names = [os.path.split(item)[1]]

        for rem, loc in zip(remote_names, local_names):
            transport.logger.debug(f"[retrieval of calc {calculation.pk}] Trying to retrieve remote item '{rem}'")
            transport.get(rem, os.path.join(folder, loc), ignore_nonexisting=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Rename `db_dbnode.type` to `db_dbnode.node_type`

This is similar to migration 5ddd24e52864

Revision ID: django_0030
Revises: django_0029

"""
from alembic import op

revision = 'django_0030'
down_revision = 'django_0029'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.alter_column('db_dbnode', 'type', new_column_name='node_type')  # pylint: disable=no-member
    # note index names are (mistakenly) not changed here


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0030.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Translator for node"""
from importlib._bootstrap import _exec, _load
import importlib.machinery
import importlib.util
import inspect
import os
import pkgutil
import sys

from aiida import orm
from aiida.common.exceptions import (
    EntryPointError,
    InputValidationError,
    InvalidOperation,
    LoadingEntryPointError,
    ValidationError,
)
from aiida.manage import get_manager
from aiida.orm import Data, Node
from aiida.plugins.entry_point import get_entry_point_names, load_entry_point
from aiida.restapi.common.exceptions import RestFeatureNotAvailable, RestInputValidationError, RestValidationError
from aiida.restapi.common.identifiers import (
    construct_full_type,
    get_full_type_filters,
    get_node_namespace,
    load_entry_point_from_full_type,
)
from aiida.restapi.translator.base import BaseTranslator


class NodeTranslator(BaseTranslator):
    # pylint: disable=too-many-instance-attributes,anomalous-backslash-in-string,too-many-arguments,too-many-branches,fixme
    """
    Translator relative to resource 'nodes' and aiida class Node
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = 'nodes'

    # The AiiDA class one-to-one associated to the present class
    _aiida_class = orm.Node

    # The string name of the AiiDA class
    _aiida_type = 'node.Node'

    # If True (False) the corresponding AiiDA class has (no) uuid property
    _has_uuid = True

    _result_type = __label__

    _content_type = None

    _attributes_filter = None
    _extras_filter = None
    _download_format = None
    _download = None
    _filename = None

    def __init__(self, **kwargs):
        """
        Initialise the parameters.
        Create the basic query_help
        """

        # basic initialization
        super().__init__(**kwargs)

        self._default_projections = ['id', 'label', 'node_type', 'process_type', 'ctime', 'mtime', 'uuid', 'user_id']

        # Inspect the subclasses of NodeTranslator, to avoid hard-coding
        # (should resemble the following tree)
        r"""
                                              /- CodeTranslator
                                             /
                                            /- KpointsTranslator
                                           /
                           /- DataTranslator -- StructureTranslator
                          /                \
                         /                  \- BandsTranslator
                        /
        NodeTranslator
                        \
                         \- CalculationTranslator
        """

        self._subclasses = self._get_subclasses()
        self._backend = get_manager().get_profile_storage()

    def set_query_type(
        self,
        query_type,
        attributes_filter=None,
        extras_filter=None,
        download_format=None,
        download=None,
        filename=None
    ):
        """
        sets one of the mutually exclusive values for self._result_type and
        self._content_type.

        :param query_type:(string) the value assigned to either variable.
        """

        if query_type == 'default':
            pass
        elif query_type == 'incoming':
            self._result_type = 'with_outgoing'
        elif query_type == 'outgoing':
            self._result_type = 'with_incoming'
        elif query_type == 'attributes':
            self._content_type = 'attributes'
            self._attributes_filter = attributes_filter
        elif query_type == 'extras':
            self._content_type = 'extras'
            self._extras_filter = extras_filter
        elif query_type == 'derived_properties':
            self._content_type = 'derived_properties'
        elif query_type == 'download':
            self._content_type = 'download'
            self._download_format = download_format
            self._download = download
        elif query_type == 'comments':
            self._content_type = 'comments'
        elif query_type == 'repo_list':
            self._content_type = 'repo_list'
            self._filename = filename
        elif query_type == 'repo_contents':
            self._content_type = 'repo_contents'
            self._filename = filename
        else:
            raise InputValidationError(f'invalid result/content value: {query_type}')

        # Add input/output relation to the query help
        if self._result_type != self.__label__:
            edge_tag = f'{self.__label__}--{self._result_type}'
            self._query_help['path'].append({
                'cls': self._aiida_class,
                'tag': self._result_type,
                'edge_tag': edge_tag,
                self._result_type: self.__label__
            })
            self._query_help['project'][edge_tag] = [{'label': {}}, {'type': {}}]

    def set_query(
        self,
        filters=None,
        orders=None,
        projections=None,
        query_type=None,
        node_id=None,
        download_format=None,
        download=None,
        filename=None,
        attributes=None,
        attributes_filter=None,
        extras=None,
        extras_filter=None,
        full_type=None
    ):
        """
        Adds filters, default projections, order specs to the query_help,
        and initializes the qb object

        :param filters: dictionary with the filters
        :param orders: dictionary with the order for each tag
        :param projections: dictionary with the projection. It is discarded
            if query_type=='attributes'/'extras'
        :param query_type: (string) specify the result or the content ("attr")
        :param id: (integer) id of a specific node
        :param download_format: file format to download e.g. cif, xyz
        :param filename: name of the file to return its content
        :param attributes: flag to show attributes for nodes
        :param attributes_filter: list of attributes to query
        :param extras: flag to show extras for nodes
        :param extras_filter: list of extras to query
        """
        # pylint: disable=arguments-differ, too-many-locals

        ## Check the compatibility of query_type and id
        if query_type != 'default' and id is None:
            raise ValidationError(
                'non default result/content can only be '
                'applied to a specific node (specify an id)'
            )

        ## Set the type of query
        self.set_query_type(
            query_type,
            attributes_filter=attributes_filter,
            extras_filter=extras_filter,
            download_format=download_format,
            download=download,
            filename=filename
        )

        ## Define projections
        if self._content_type is not None:
            # Use '*' so that the object itself will be returned.
            # In get_results() we access attributes/extras by
            # calling the attributes/extras.
            projections = ['*']
        else:
            pass  # i.e. use the input parameter projection

        # TODO this actually works, but the logic is a little bit obscure.
        # Make it clearer
        if self._result_type is not self.__label__:
            projections = self._default_projections

        if full_type:
            # If full_type is not none, convert it into node_type
            # and process_type filters to pass it to query help
            full_type_filter = get_full_type_filters(full_type)
            filters.update(full_type_filter)

        super().set_query(
            filters=filters,
            orders=orders,
            projections=projections,
            node_id=node_id,
            attributes=attributes,
            attributes_filter=attributes_filter,
            extras=extras,
            extras_filter=extras_filter
        )

    def _get_content(self):
        """
        Used by get_results() in case of endpoint include "content" option
        :return: data: a dictionary containing the results obtained by
        running the query
        """
        # pylint: disable=too-many-statements

        if not self._is_qb_initialized:
            raise InvalidOperation('query builder object has not been initialized.')

        # Count the total number of rows returned by the query (if not already done)
        if self._total_count is None:
            self.count()

        # If count==0 return
        if self._total_count == 0:
            return {}

        # otherwise ...
        node = self.qbobj.first()[0]  # pylint: disable=unsubscriptable-object

        # content/attributes
        if self._content_type == 'attributes':
            # Get all attrs if attributes_filter is None
            if self._attributes_filter is None:
                data = {self._content_type: node.attributes}
            # Get all attrs contained in attributes_filter
            else:
                attrs = {}
                for key in node.attributes.keys():
                    if key in self._attributes_filter:
                        attrs[key] = node.get_attribute(key)
                data = {self._content_type: attrs}

        # content/extras
        elif self._content_type == 'extras':
            # Get all extras if elist is None
            if self._extras_filter is None:
                data = {self._content_type: node.extras}
            else:
                # Get all extras contained in elist
                extras = {}
                for key in node.extras.keys():
                    if key in self._extras_filter:
                        extras[key] = node.get_extra(key)
                data = {self._content_type: extras}

        # Data needed for visualization appropriately serialized (this
        # actually works only for data derived classes)
        # TODO refactor the code so to have this option only in data and
        # derived classes
        elif self._content_type == 'derived_properties':
            data = {self._content_type: self.get_derived_properties(node)}

        elif self._content_type == 'download':
            # In this we do not return a dictionary but download the file in
            # specified format if available
            data = {self._content_type: self.get_downloadable_data(node, self._download_format)}

        elif self._content_type == 'repo_list':
            # return list of all the files and directories from node file repository
            data = {self._content_type: self.get_repo_list(node, self._filename)}

        elif self._content_type == 'repo_contents':
            # return the contents of single file from node file repository
            data = self.get_repo_contents(node, self._filename)

        elif self._content_type == 'comments':
            # return the node comments
            data = {self._content_type: self.get_comments(node)}

        else:
            raise ValidationError('invalid content type')

        return data

    def _get_subclasses(self, parent=None, parent_class=None, recursive=True):
        """
        Import all submodules of the package containing the present class.
        Includes subpackages recursively, if specified.

        :param parent: package/class.
            If package looks for the classes in submodules.
            If class, first looks for the package where it is contained
        :param parent_class: class of which to look for subclasses
        :param recursive: True/False (go recursively into submodules)
        """

        # If no parent class is specified set it to self.__class
        parent = self.__class__ if parent is None else parent

        # Suppose parent is class
        if inspect.isclass(parent):

            # Set the package where the class is contained
            classfile = inspect.getfile(parent)
            package_path = os.path.dirname(classfile)

            # If parent class is not specified, assume it is the parent
            if parent_class is None:
                parent_class = parent

        # Suppose parent is a package (directory containing __init__.py).
        # Check if it contains attribute __path__
        elif inspect.ismodule(parent) and hasattr(parent, '__path__'):

            # Assume path is one element list
            package_path = parent.__path__[0]

            # if parent is a package, parent_class cannot be None
            if parent_class is None:
                raise TypeError('argument parent_class cannot be None')

        # Recursively check the subpackages
        results = {}

        for _, name, is_pkg in pkgutil.walk_packages([package_path]):
            # N.B. pkgutil.walk_packages requires a LIST of paths

            full_path_base = os.path.join(package_path, name)
            if is_pkg:
                # re-implementation of deprecated `imp.load_package`
                if os.path.isdir(full_path_base):
                    #Adds an extension to check for __init__ file in the package directory
                    extensions = (importlib.machinery.SOURCE_SUFFIXES[:] + importlib.machinery.BYTECODE_SUFFIXES[:])
                    for extension in extensions:
                        init_path = os.path.join(full_path_base, '__init__' + extension)
                        if os.path.exists(init_path):
                            path = init_path
                            break
                    else:
                        raise ValueError(f'{full_path_base!r} is not a package')
                #passing [] to submodule_search_locations indicates its a package and python searches for sub-modules
                spec = importlib.util.spec_from_file_location(full_path_base, path, submodule_search_locations=[])
                if full_path_base in sys.modules:
                    #executes from sys.modules
                    app_module = _exec(spec, sys.modules[full_path_base])
                else:
                    #loads and executes the module
                    app_module = _load(spec)
            else:
                full_path = f'{full_path_base}.py'
                # reimplementation of deprecated `imp.load_source`
                spec = importlib.util.spec_from_file_location(name, full_path)
                app_module = importlib.util.module_from_spec(spec)
                sys.modules[name] = app_module
                spec.loader.exec_module(app_module)

            # Go through the content of the module
            if not is_pkg:
                for fname, obj in inspect.getmembers(app_module):
                    if inspect.isclass(obj) and issubclass(obj, parent_class):
                        results[fname] = obj
            # Look in submodules
            elif is_pkg and recursive:
                results.update(self._get_subclasses(parent=app_module, parent_class=parent_class))

        return results

    def get_derived_properties(self, node):
        """
        Generic function to get the derived properties of the node.
        Actual definition is in child classes as the content to be
        returned depends on the plugin specific
        to the resource

        :param node: node object that has to be visualized
        :returns: derived properties of the node

        If this method is called by Node resource it will look for the type
        of object and invoke the correct method in the lowest-compatible
        subclass
        """

        # Look for the translator associated to the class of which this node
        # is instance
        tclass = type(node)

        for subclass in self._subclasses.values():
            if subclass._aiida_type.split('.')[-1] == tclass.__name__:  # pylint: disable=protected-access
                lowtrans = subclass

        derived_properties = lowtrans.get_derived_properties(node)

        return derived_properties

    @staticmethod
    def get_all_download_formats(full_type=None):
        """
        returns dict of possible node formats for all available node types
        """

        all_formats = {}

        if full_type:
            try:
                node_cls = load_entry_point_from_full_type(full_type)
            except (TypeError, ValueError):
                raise RestInputValidationError(f'The full type {full_type} is invalid.')
            except EntryPointError:
                raise RestFeatureNotAvailable('The download formats for this node type are not available.')

            try:
                available_formats = node_cls.get_export_formats()
                all_formats[full_type] = available_formats
            except AttributeError:
                pass
        else:
            entry_point_group = 'aiida.data'

            for name in get_entry_point_names(entry_point_group):
                try:
                    node_cls = load_entry_point(entry_point_group, name)
                    available_formats = node_cls.get_export_formats()
                except (AttributeError, LoadingEntryPointError):
                    continue

                if available_formats:
                    full_type = construct_full_type(node_cls.class_node_type, '')
                    all_formats[full_type] = available_formats

        return all_formats

    @staticmethod
    def get_downloadable_data(node, download_format=None):
        """
        Generic function to download file in specified format.
        Actual definition is in child classes as the content to be
        returned and its format depends on the download plugin specific
        to the resource

        :param node: node object
        :param download_format: file extension format
        :returns: data in selected format to download

        If this method is called for a Data node resource it will
        invoke the get_downloadable_data method in the Data transaltor.
        Otherwise it raises RestFeatureNotAvailable exception
        """

        # This needs to be here because currently, for all nodes the `NodeTranslator` will be instantiated. Once that
        # logic is cleaned where the correct translator sub class is instantiated based on the node type that is
        # referenced, this hack can be removed.
        if isinstance(node, Data):
            from .data import DataTranslator  # pylint: disable=cyclic-import
            downloadable_data = DataTranslator.get_downloadable_data(node, download_format=download_format)
            return downloadable_data

        raise RestFeatureNotAvailable('This endpoint is not available for Process nodes.')

    @staticmethod
    def get_repo_list(node, filename=''):
        """
        Every node in AiiDA is having repo folder.
        This function returns the metadata using list_objects() method
        :param node: node object
        :param filename: folder name (optional)
        :return: folder list
        """
        try:
            flist = node.list_objects(filename)
        except NotADirectoryError:
            raise RestInputValidationError(f'{filename} is not a directory in this repository')
        response = []
        for fobj in flist:
            response.append({'name': fobj.name, 'type': fobj.file_type.name})
        return response

    @staticmethod
    def get_repo_contents(node, filename=''):
        """
        Every node in AiiDA is having repo folder.
        This function returns the metadata using get_object() method
        :param node: node object
        :param filename: folder or file name (optional)
        :return: file content in bytes to download
        """

        if filename:
            try:
                data = node.get_object_content(filename, mode='rb')
                return data
            except FileNotFoundError:
                raise RestInputValidationError('No such file is present')
        raise RestValidationError('filename is not provided')

    @staticmethod
    def get_comments(node):
        """
        :param node: node object
        :return: node comments
        """
        comments = node.get_comments()
        response = []
        for cobj in comments:
            response.append({
                'created_time': cobj.ctime,
                'modified_time': cobj.mtime,
                'user': f'{cobj.user.first_name} {cobj.user.last_name}',
                'message': cobj.content
            })
        return response

    @staticmethod
    def get_file_content(node, file_name):
        """
        It reads the file from directory and returns its content.

        Instead of using "send_from_directory" from flask, this method is written
        because in next aiida releases the file can be stored locally or in object storage.

        :param node: aiida folderData node which contains file
        :param file_name: name of the file to return its contents
        :return:
        """
        file_parts = file_name.split(os.sep)

        if len(file_parts) > 1:
            file_name = file_parts.pop()
            for folder in file_parts:
                node = node.get_subfolder(folder)

        with node.open(file_name) as fobj:
            return fobj.read()

    def get_results(self):
        """
        Returns either a list of nodes or details of single node from database

        :return: either a list of nodes or the details of single node
            from the database
        """
        if self._content_type is not None:
            return self._get_content()

        return super().get_results()

    def get_formatted_result(self, label):
        """
        Runs the query and retrieves results tagged as "label".

        :param label: the tag of the results to be extracted out of
          the query rows.
        :type label: str
        :return: a list of the query results
        """

        results = super().get_formatted_result(label)

        if self._result_type == 'with_outgoing':
            result_name = 'incoming'
        elif self._result_type == 'with_incoming':
            result_name = 'outgoing'
        else:
            result_name = self.__label__

        for node_entry in results[result_name]:
            # construct full_type and add it to every node
            node_entry['full_type'] = (
                construct_full_type(node_entry.get('node_type'), node_entry.get('process_type'))
                if node_entry.get('node_type') or node_entry.get('process_type') else None
            )

        return results

    def get_statistics(self, user_pk=None):
        """Return statistics for a given node"""
        return self._backend.query().get_creation_statistics(user_pk=user_pk)

    @staticmethod
    def get_namespace(user_pk=None, count_nodes=False):
        """
        return full_types of the nodes
        """

        return get_node_namespace(user_pk=user_pk, count_nodes=count_nodes).get_description()

    def get_io_tree(self, uuid_pattern, tree_in_limit, tree_out_limit):
        # pylint: disable=too-many-statements,too-many-locals
        """
        json data to display nodes in tree format
        :param uuid_pattern: main node uuid
        :return: json data to display node tree
        """

        def get_node_description(node):
            """
            Get the description of the node.
            CalcJobNodes migrated from AiiDA < 1.0.0 do not have a valid CalcJobState,
            in this case the function returns as description the type of the node (CalcJobNode)
            :param node: node object
            :return: description of the node
            """
            try:
                description = node.get_description()
            except ValueError:
                description = node.node_type.split('.')[-2]
            return description

        # Check whether uuid_pattern identifies a unique node
        self._check_id_validity(uuid_pattern)

        qb_obj = orm.QueryBuilder()
        qb_obj.append(Node, tag='main', project=['*'], filters=self._id_filter)

        nodes = []

        if qb_obj.count() > 0:
            main_node = qb_obj.first(flat=True)
            pk = main_node.pk
            uuid = main_node.uuid
            nodetype = main_node.node_type
            nodelabel = main_node.label
            description = get_node_description(main_node)
            ctime = main_node.ctime
            mtime = main_node.mtime

            nodes.append({
                'ctime': ctime,
                'mtime': mtime,
                'id': pk,
                'uuid': uuid,
                'node_type': nodetype,
                'node_label': nodelabel,
                'description': description,
                'incoming': [],
                'outgoing': []
            })

        # get all incoming
        qb_obj = orm.QueryBuilder()
        qb_obj.append(Node, tag='main', project=['*'], filters=self._id_filter)
        qb_obj.append(Node, tag='in', project=['*'], edge_project=['label', 'type'],
                      with_outgoing='main').order_by({'in': [{
                          'id': {
                              'order': 'asc'
                          }
                      }]})
        if tree_in_limit is not None:
            qb_obj.limit(tree_in_limit)

        sent_no_of_incomings = qb_obj.count()

        if sent_no_of_incomings > 0:
            for node_input in qb_obj.iterdict():
                node = node_input['in']['*']
                pk = node.pk
                linklabel = node_input['main--in']['label']
                linktype = node_input['main--in']['type']
                uuid = node.uuid
                nodetype = node.node_type
                nodelabel = node.label
                description = get_node_description(node)
                node_ctime = node.ctime
                node_mtime = node.mtime

                nodes[0]['incoming'].append({
                    'ctime': node_ctime,
                    'mtime': node_mtime,
                    'id': pk,
                    'uuid': uuid,
                    'node_type': nodetype,
                    'node_label': nodelabel,
                    'description': description,
                    'link_label': linklabel,
                    'link_type': linktype
                })

        # get all outgoing
        qb_obj = orm.QueryBuilder()
        qb_obj.append(Node, tag='main', project=['*'], filters=self._id_filter)
        qb_obj.append(Node, tag='out', project=['*'], edge_project=['label', 'type'],
                      with_incoming='main').order_by({'out': [{
                          'id': {
                              'order': 'asc'
                          }
                      }]})
        if tree_out_limit is not None:
            qb_obj.limit(tree_out_limit)

        sent_no_of_outgoings = qb_obj.count()

        if sent_no_of_outgoings > 0:
            for output in qb_obj.iterdict():
                node = output['out']['*']
                pk = node.pk
                linklabel = output['main--out']['label']
                linktype = output['main--out']['type']
                uuid = node.uuid
                nodetype = node.node_type
                nodelabel = node.label
                description = get_node_description(node)
                node_ctime = node.ctime
                node_mtime = node.mtime

                nodes[0]['outgoing'].append({
                    'ctime': node_ctime,
                    'mtime': node_mtime,
                    'id': pk,
                    'uuid': uuid,
                    'node_type': nodetype,
                    'node_label': nodelabel,
                    'description': description,
                    'link_label': linklabel,
                    'link_type': linktype
                })

        # count total no of nodes
        builder = orm.QueryBuilder()
        builder.append(Node, tag='main', project=['id'], filters=self._id_filter)
        builder.append(Node, tag='in', project=['id'], with_outgoing='main')
        total_no_of_incomings = builder.count()

        builder = orm.QueryBuilder()
        builder.append(Node, tag='main', project=['id'], filters=self._id_filter)
        builder.append(Node, tag='out', project=['id'], with_incoming='main')
        total_no_of_outgoings = builder.count()

        metadata = [{
            'total_no_of_incomings': total_no_of_incomings,
            'total_no_of_outgoings': total_no_of_outgoings,
            'sent_no_of_incomings': sent_no_of_incomings,
            'sent_no_of_outgoings': sent_no_of_outgoings
        }]

        return {'nodes': nodes, 'metadata': metadata}

    def get_projectable_properties(self):
        """
        Get projectable properties specific for Node
        :return: dict of projectable properties and column_order list
        """
        projectable_properties = {
            'creator': {
                'display_name': 'Creator',
                'help_text': 'User that created the node',
                'is_foreign_key': False,
                'type': 'str',
                'is_display': True
            },
            'ctime': {
                'display_name': 'Creation time',
                'help_text': 'Creation time of the node',
                'is_foreign_key': False,
                'type': 'datetime.datetime',
                'is_display': True
            },
            'label': {
                'display_name': 'Label',
                'help_text': 'User-assigned label',
                'is_foreign_key': False,
                'type': 'str',
                'is_display': False
            },
            'mtime': {
                'display_name': 'Last Modification time',
                'help_text': 'Last modification time',
                'is_foreign_key': False,
                'type': 'datetime.datetime',
                'is_display': True
            },
            'node_type': {
                'display_name': 'Type',
                'help_text': 'Node type',
                'is_foreign_key': False,
                'type': 'str',
                'is_display': True
            },
            'process_type': {
                'display_name': 'Process type',
                'help_text': 'Process type',
                'is_foreign_key': False,
                'type': 'str',
                'is_display': False
            },
            'user_id': {
                'display_name': 'Id of creator',
                'help_text': 'Id of the user that created the node',
                'is_foreign_key': True,
                'related_column': 'id',
                'related_resource': 'users',
                'type': 'int',
                'is_display': False
            },
            'uuid': {
                'display_name': 'Unique ID',
                'help_text': 'Universally Unique Identifier',
                'is_foreign_key': False,
                'type': 'unicode',
                'is_display': True
            }
        }

        # Note: final schema will contain details for only the fields present in column order
        column_order = ['uuid', 'label', 'node_type', 'ctime', 'mtime', 'creator']

        return projectable_properties, column_order



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=no-self-use
""""Implementation of `DbImporter` for the CISD database."""
import io

from aiida.tools.dbimporters.baseclasses import CifEntry, DbImporter, DbSearchResults


class IcsdImporterExp(Exception):
    """Base class for ICSD exceptions."""


class CifFileErrorExp(IcsdImporterExp):
    """Raised when the author loop is missing in a CIF file."""


class NoResultsWebExp(IcsdImporterExp):
    """Raised when a webpage query returns no results."""


class IcsdDbImporter(DbImporter):
    """
    Importer for the Inorganic Crystal Structure Database, short ICSD, provided by
    FIZ Karlsruhe. It allows to run queries and analyse all the results.

    :param server: Server URL, the web page of the database. It is
        required in order to have access to the full database.
        I t should contain both the protocol and the domain name
        and end with a slash, as in::

          server = "http://ICSDSERVER.com/"

    :param urladd: part of URL which is added between query and and the server URL
        (default: ``index.php?``). only needed for web page query
    :param querydb: boolean, decides whether the mysql database is queried
        (default: True).
        If False, the query results are obtained through the web page
        query, which is
        restricted to a maximum of 1000 results per query.
    :param dl_db: icsd comes with a full (default: ``icsd``) and a demo
        database (``icsdd``).
        This parameter allows the user to switch to the demo database
        for testing purposes, if the access rights to the full database
        are not granted.
    :param host: MySQL database host. If the MySQL database is hosted on
        a different machine, use  "127.0.0.1" as host, and open
        a SSH tunnel to the host using::

            ssh -L 3306:localhost:3306 username@hostname.com

        or (if e.g. you get an URLError with Errno 111 (Connection refused)
        upon querying)::

            ssh -L 3306:localhost:3306 -L 8010:localhost:80 username@hostname.com

    :param user: mysql database username (default: dba)
    :param passwd: mysql database password (default: sql)
    :param db: name of the database (default: icsd)
    :param port: Port to access the mysql database (default: 3306)
    """

    length_precision = 0.001
    angle_precision = 0.001
    volume_precision = 0.001
    temperature_precision = 0.001
    density_precision = 0.001
    pressure_precision = 1

    def __init__(self, **kwargs):

        self.db_parameters = {
            'server': '',
            'urladd': 'index.php?',
            'querydb': True,
            'dl_db': 'icsd',
            'host': '',
            'user': 'dba',
            'passwd': 'sql',
            'db': 'icsd',
            'port': '3306',
        }
        self.setup_db(**kwargs)

    # for mysql db query
    def _int_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying integer fields
        :param key: Database keyword
        :param alias: Query parameter name
        :param values: Corresponding values from query
        :return: SQL query predicate
        """
        for value in values:
            if not isinstance(value, int) and not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only integers and strings are accepted')
        return f"{key} IN ({', '.join(str(int(i)) for i in values)})"

    def _str_exact_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying string fields.
        """
        for value in values:
            if not isinstance(value, int) and not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only integers and strings are accepted')
        return '{} IN ({})'.format(key, ', '.join("'{}'".format(f) for f in values))

    def _formula_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying formula fields.
        """
        for value in values:
            if not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only strings are accepted')
        return self._str_exact_clause(key, alias, [str(f) for f in values])

    def _str_fuzzy_clause(self, key, alias, values):
        """
        Return SQL query predicate for fuzzy querying of string fields.
        """
        for value in values:
            if not isinstance(value, int) and not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only integers and strings are accepted')
        return ' OR '.join(f"{key} LIKE '%{s}%'" for s in values)

    def _composition_clause(self, key, alias, values):  # pylint: disable=unused-argument
        """
        Return SQL query predicate for querying elements in formula fields.
        """
        for value in values:
            if not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only strings are accepted')
        # SUM_FORM in the ICSD always stores a numeral after the element name,
        # STRUCT_FORM does not, so it's better to use SUM_FORM for the composition query.
        # The element-numeral pair can be in the beginning of the formula expression (therefore no space before),
        # or at the end of the formula expression (no space after).
        # Be aware that one needs to check that space/beginning of line before and ideally also space/end of line
        # after, because I found that capitalization of the element name is not enforced in these queries.
        return ' AND '.join(r'SUM_FORM REGEXP \'(^|\ ){}[0-9\.]+($|\ )\''.format(value) for value in values)

    def _double_clause(self, key, alias, values, precision):
        """
        Return SQL query predicate for querying double-valued fields.
        """
        for value in values:
            if not isinstance(value, int) and not isinstance(value, float):
                raise ValueError("incorrect value for keyword '" + alias + ' only integers and floats are accepted')
        return ' OR '.join(f'{key} BETWEEN {d - precision} AND {d + precision}' for d in values)

    def _crystal_system_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying crystal_system.
        """
        valid_systems = {
            'cubic': 'CU',
            'hexagonal': 'HE',
            'monoclinic': 'MO',
            'orthorhombic': 'OR',
            'tetragonal': 'TE',
            'trigonal': 'TG',
            'triclinic': 'TC'
        }  # from icsd accepted crystal systems

        for value in values:
            if not isinstance(value, int) and not isinstance(value, str):
                raise ValueError("incorrect value for keyword '" + alias + ' only strings are accepted')
        return key + ' IN (' + ', '.join("'" + valid_systems[f.lower()] + "'" for f in values) + ')'

    def _length_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying lattice vector lengths.
        """
        return self._double_clause(key, alias, values, self.length_precision)

    def _density_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying density.
        """
        return self._double_clause(key, alias, values, self.density_precision)

    def _angle_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying lattice angles.
        """
        return self._double_clause(key, alias, values, self.angle_precision)

    def _volume_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying unit cell volume.
        """
        return self._double_clause(key, alias, values, self.volume_precision)

    def _temperature_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying temperature.
        """
        return self._double_clause(key, alias, values, self.temperature_precision)

    def _pressure_clause(self, key, alias, values):
        """
        Return SQL query predicate for querying pressure.
        """
        return self._double_clause(key, alias, values, self.pressure_precision)

    # for the web query
    @staticmethod
    def _parse_all(key, values):  # pylint: disable=unused-argument
        """
        Convert numbers, strings, lists into strings.
        :param key: query parameter
        :param values: corresponding values
        :return retval: string
        """
        if isinstance(values, list):
            retval = ' '.join(values)
        elif isinstance(values, int):
            retval = str(values)
        elif isinstance(values, str):
            retval = values
        return retval

    @staticmethod
    def _parse_number(key, values):  # pylint: disable=unused-argument
        """
        Convert int into string.
        :param key: query parameter
        :param values: corresponding values
        :return retval: string
        """
        if isinstance(values, int):
            retval = str(values)
        elif isinstance(values, str):
            retval = values
        return retval

    @staticmethod
    def _parse_mineral(key, values):
        """
        Convert mineral_name and chemical_name into right format.
        :param key: query parameter
        :param values: corresponding values
        :return retval: string
        """
        if key == 'mineral_name':
            retval = 'M=' + values
        elif key == 'chemical_name':
            retval = 'C=' + values
        return retval

    @staticmethod
    def _parse_volume(key, values):  # pylint: disable=too-many-return-statements
        """
        Convert volume, cell parameter and angle queries into right format.
        :param key: query parameter
        :param values: corresponding values
        :return retval: string
        """
        if key == 'volume':
            return 'v=' + values

        if key == 'a':
            return 'a=' + values

        if key == 'b':
            return 'b=' + values

        if key == 'c':
            return 'c=' + values

        if key == 'alpha':
            return 'al=' + values

        if key == 'beta':
            return 'be=' + values

        if key == 'gamma':
            return 'ga=' + values

    @staticmethod
    def _parse_system(key, values):  # pylint: disable=unused-argument
        """
        Return crystal system in the right format.
        :param key: query parameter
        :param values: corresponding values
        :return retval: string
        """
        valid_systems = {
            'cubic': 'CU',
            'hexagonal': 'HE',
            'monoclinic': 'MO',
            'orthorhombic': 'OR',
            'tetragonal': 'TE',
            'trigonal': 'TG',
            'triclinic': 'TC'
        }

        return valid_systems[values.lower()]

    # mysql database - query parameter (alias) : [mysql keyword (key), function to call]
    keywords_db = {
        'id': ['COLL_CODE', _int_clause],
        'element': ['SUM_FORM;', _composition_clause],
        'number_of_elements': ['EL_COUNT', _int_clause],
        'chemical_name': ['CHEM_NAME', _str_fuzzy_clause],
        'formula': ['SUM_FORM', _formula_clause],
        'volume': ['C_VOL', _volume_clause],
        'spacegroup': ['SGR', _str_exact_clause],
        'a': ['A_LEN', _length_clause],
        'b': ['B_LEN', _length_clause],
        'c': ['C_LEN', _length_clause],
        'alpha': ['ALPHA', _angle_clause],
        'beta': ['BETA', _angle_clause],
        'gamma': ['GAMMA', _angle_clause],
        'density': ['DENSITY_CALC', _density_clause],
        'wyckoff': ['WYCK', _str_exact_clause],
        'molar_mass': ['MOL_MASS', _density_clause],
        'pdf_num': ['PDF_NUM', _str_exact_clause],
        'z': ['Z', _int_clause],
        'measurement_temp': ['TEMPERATURE', _temperature_clause],
        'authors': ['AUTHORS_TEXT', _str_fuzzy_clause],
        'journal': ['journal', _str_fuzzy_clause],
        'title': ['AU_TITLE', _str_fuzzy_clause],
        'year': ['MPY', _int_clause],
        'crystal_system': ['CRYST_SYS_CODE', _crystal_system_clause],
    }
    # keywords accepted for the web page query
    keywords = {
        'id': ('authors', _parse_all),
        'authors': ('authors', _parse_all),
        'element': ('elements', _parse_all),
        'number_of_elements': ('elementc', _parse_all),
        'mineral_name': ('mineral', _parse_mineral),
        'chemical_name': ('mineral', _parse_mineral),
        'formula': ('formula', _parse_all),
        'volume': ('volume', _parse_volume),
        'a': ('volume', _parse_volume),
        'b': ('volume', _parse_volume),
        'c': ('volume', _parse_volume),
        'alpha': ('volume', _parse_volume),
        'beta': ('volume', _parse_volume),
        'gamma': ('volume', _parse_volume),
        'spacegroup': ('spaceg', _parse_all),
        'journal': ('journal', _parse_all),
        'title': ('title', _parse_all),
        'year': ('year', _parse_all),
        'crystal_system': ('system', _parse_system),
    }

    def query(self, **kwargs):
        """
        Depending on the db_parameters, the mysql database or the web page are queried.
        Valid parameters are found using IcsdDbImporter.get_supported_keywords().

        :param kwargs: A list of ''keyword = [values]'' pairs.
        """
        if self.db_parameters['querydb']:
            return self._query_sql_db(**kwargs)

        return self._queryweb(**kwargs)

    def _query_sql_db(self, **kwargs):
        """
        Perform a query on Icsd mysql database using ``keyword = value`` pairs,
        specified in ``kwargs``. Returns an instance of IcsdSearchResults.
        :param kwargs: A list of ``keyword = [values]`` pairs
        :return: IcsdSearchResults
        """

        sql_where_query = []  # second part of sql query

        for key, value in kwargs.items():
            if not isinstance(value, list):
                value = [value]
            sql_where_query.append(f'({self.keywords_db[key][1](self, self.keywords_db[key][0], key, value)})')
        if 'crystal_system' in kwargs:  # to query another table than the main one, add LEFT JOIN in front of WHERE
            sql_query = 'LEFT JOIN space_group ON space_group.sgr=icsd.sgr LEFT '\
                        'JOIN space_group_number ON '\
                        'space_group_number.sgr_num=space_group.sgr_num '\
                        + 'WHERE' + ' AND '.join(sql_where_query)
        elif sql_where_query:
            sql_query = 'WHERE' + ' AND '.join(sql_where_query)
        else:
            sql_query = ''

        return IcsdSearchResults(query=sql_query, db_parameters=self.db_parameters)

    def _queryweb(self, **kwargs):
        """
        Perform a query on the Icsd web database using ``keyword = value`` pairs,
        specified in ``kwargs``. Returns an instance of IcsdSearchResults.
        :note: Web search has a maximum result number fixed at 1000.
        :param kwargs: A list of ``keyword = [values]`` pairs
        :return: IcsdSearchResults
        """
        from urllib.parse import urlencode
        self.actual_args = {  # pylint: disable=attribute-defined-outside-init
            'action': 'Search',
            'nb_rows': '100',  # max is 100
            'order_by': 'yearDesc',
            'authors': '',
            'volume': '',
            'mineral': ''
        }

        for key, value in kwargs.items():
            try:
                realname = self.keywords[key][0]
                newv = self.keywords[key][1](key, value)
                # Because different keys correspond to the same search field.
                if realname in ['authors', 'volume', 'mineral']:
                    self.actual_args[realname] = self.actual_args[realname] + newv + ' '
                else:
                    self.actual_args[realname] = newv
            except KeyError as exc:
                raise TypeError(f"ICSDImporter got an unexpected keyword argument '{exc.args[0]}'")

        url_values = urlencode(self.actual_args)
        query_url = self.db_parameters['urladd'] + url_values

        return IcsdSearchResults(query=query_url, db_parameters=self.db_parameters)

    def setup_db(self, **kwargs):
        """
        Change the database connection details.
        At least the host server has to be defined.

        :param kwargs: db_parameters for the mysql database connection
          (host, user, passwd, db, port)
        """
        for key in self.db_parameters:
            if key in kwargs:
                self.db_parameters[key] = kwargs[key]

    def get_supported_keywords(self):
        """
        :return: List of all supported query keywords.
        """
        if self.db_parameters['querydb']:
            return self.keywords_db.keys()

        return self.keywords.keys()


class IcsdSearchResults(DbSearchResults):  # pylint: disable=abstract-method,too-many-instance-attributes
    """
    Result manager for the query performed on ICSD.

    :param query: mysql query or webpage query
    :param db_parameters: database parameter setup during the
      initialisation of the IcsdDbImporter.
    """
    cif_url = '/index.php?format=cif&action=Export&id%5B%5D={}'
    db_name = 'Icsd'

    def __init__(self, query, db_parameters):
        # pylint: disable=super-init-not-called
        self.db = None  # pylint: disable=invalid-name
        self.cursor = None
        self.db_parameters = db_parameters
        self.query = query
        self.number_of_results = None
        self._results = []
        self.cif_numbers = []
        self.entries = {}
        self.page = 1
        self.position = 0
        self.db_version = None
        self.sql_select_query = 'SELECT SQL_CALC_FOUND_ROWS icsd.IDNUM, icsd.COLL_CODE, icsd.STRUCT_FORM '
        self.sql_from_query = 'FROM icsd '

        if self.db_parameters['querydb']:
            self.query_db_version()
        self.query_page()

    @property
    def results(self):
        """
        Return the list of results
        """
        return self._results

    def next(self):
        """
        Return next result as IcsdEntry.
        """
        if self.number_of_results > self.position:
            self.position = self.position + 1
            return self.at(self.position - 1)

        self.position = 0
        raise StopIteration()

    def at(self, position):
        """
        Return ``position``-th result as IcsdEntry.
        """

        if position < 0 or position >= self.number_of_results:
            raise IndexError('index out of bounds')
        while position + 1 >= len(self._results) and len(self._results) < self.number_of_results:
            self.page = self.page + 1
            self.query_page()

        if position not in self.entries:
            if self.db_parameters['querydb']:
                self.entries[position] = IcsdEntry(
                    self.db_parameters['server'] + self.db_parameters['dl_db'] +
                    self.cif_url.format(self._results[position]),
                    db_name=self.db_name,
                    id=self.cif_numbers[position],
                    version=self.db_version,
                    extras={'idnum': self._results[position]}
                )
            else:
                self.entries[position] = IcsdEntry(
                    self.db_parameters['server'] + self.db_parameters['dl_db'] +
                    self.cif_url.format(self._results[position]),
                    db_name=self.db_name,
                    extras={'idnum': self._results[position]}
                )
        return self.entries[position]

    def query_db_version(self):
        """
        Query the version of the icsd database (last row of RELEASE_TAGS).
        """
        results = []
        if self.db_parameters['querydb']:

            sql_select_query = 'SELECT RELEASE_TAG '
            sql_from_query = 'FROM icsd.icsd_database_information '

            self._connect_db()
            query_statement = f'{sql_select_query}{sql_from_query}'
            self.cursor.execute(query_statement)
            self.db.commit()

            for row in self.cursor.fetchall():
                results.append(str(row[0]))

            self._disconnect_db()
            try:
                self.db_version = results[-1]
            except IndexError:
                raise IcsdImporterExp('Database version not found')

        else:
            raise NotImplementedError('Cannot query the database version with a web query.')

    def query_page(self):
        """
        Query the mysql or web page database, depending on the db_parameters.
        Store the number_of_results, cif file number and the corresponding icsd number.

        :note: Icsd uses its own number system, different from the CIF
                file numbers.
        """
        if self.db_parameters['querydb']:

            self._connect_db()
            query_statement = '{}{}{} LIMIT {}, 100'.format(
                self.sql_select_query, self.sql_from_query, self.query, (self.page - 1) * 100
            )
            self.cursor.execute(query_statement)
            self.db.commit()

            for row in self.cursor.fetchall():
                self._results.append(str(row[0]))
                self.cif_numbers.append(str(row[1]))

            if self.number_of_results is None:
                self.cursor.execute('SELECT FOUND_ROWS()')
                self.number_of_results = int(self.cursor.fetchone()[0])

            self._disconnect_db()

        else:
            import re
            from urllib.request import urlopen

            from bs4 import BeautifulSoup  # pylint: disable=import-error

            with urlopen(
                self.db_parameters['server'] + self.db_parameters['db'] + '/' + self.query.format(str(self.page))
            ) as handle:
                self.html = handle.read()

            self.soup = BeautifulSoup(self.html)

            try:

                if self.number_of_results is None:
                    self.number_of_results = int(re.findall(r'\d+', str(self.soup.find_all('i')[-1]))[0])
            except IndexError:
                raise NoResultsWebExp

            for i in self.soup.find_all('input', type='checkbox'):
                self._results.append(i['id'])

    def _connect_db(self):
        """
        Connect to the MySQL database for performing searches.
        """
        try:
            import MySQLdb
        except ImportError:
            import pymysql as MySQLdb

        self.db = MySQLdb.connect(
            host=self.db_parameters['host'],
            user=self.db_parameters['user'],
            passwd=self.db_parameters['passwd'],
            db=self.db_parameters['db'],
            port=int(self.db_parameters['port'])
        )
        self.cursor = self.db.cursor()

    def _disconnect_db(self):
        """
        Close connection to the MySQL database.
        """
        self.db.close()


class IcsdEntry(CifEntry):  # pylint: disable=abstract-method
    """
    Represent an entry from Icsd.

    :note:
      - Before July 2nd 2015, source['id'] contained icsd.IDNUM (internal
        icsd id number) and source['extras']['cif_nr'] the cif number
        (icsd.COLL_CODE).
      - After July 2nd 2015, source['id'] has been replaced by the cif
        number and source['extras']['idnum'] is icsd.IDNUM .
    """
    _license = 'ICSD'

    def __init__(self, uri, **kwargs):
        """
        Create an instance of IcsdEntry, related to the supplied URI.
        """
        super().__init__(**kwargs)
        self.source = {
            'db_name': kwargs.get('db_name', 'Icsd'),
            'db_uri': None,
            'id': kwargs.get('id', None),
            'version': kwargs.get('version', None),
            'uri': uri,
            'extras': {
                'idnum': kwargs.get('extras', {}).get('idnum', None)
            },
            'license': self._license,
        }

    @property
    def contents(self):
        """
        Returns raw contents of a file as string. This overrides the DbEntry implementation because
        the ICSD php backend returns the contents of the CIF in ISO-8859-1 encoding. However, the
        PyCifRW library (and most other sensible applications), expects UTF-8. Therefore, we decode
        the original CIF data to unicode and encode it in the UTF-8 format
        """
        import urllib.request

        if self._contents is None:
            from hashlib import md5
            with urllib.request.urlopen(self.source['uri']) as handle:
                self._contents = handle.read()
            self._contents = self._contents.decode('iso-8859-1').encode('utf8')
            self.source['source_md5'] = md5(self._contents).hexdigest()

        return self._contents

    def get_ase_structure(self):
        """
        :return: ASE structure corresponding to the cif file.
        """
        from aiida.orm import CifData
        cif = correct_cif(self.cif)
        return CifData.read_cif(io.StringIO(cif))


def correct_cif(cif):
    """
    Correct the format of the CIF files.
    At the moment, it only fixes missing quotes in the authors field
    (``ase.read.io`` only works if the author names are quoted,
    if not an AssertionError is raised).

    :param cif: A string containing the content of the CIF file.
    :return: a string containing the corrected CIF file.
    """
    # Do more checks to be sure it's working in everycase
    # -> no _publ_author_name, several lines, correct input
    lines = cif.split('\n')

    try:
        author_index = lines.index('_publ_author_name')
    except ValueError:
        raise CifFileErrorExp('_publ_author_name line missing in cif file')
    else:
        inc = 1
        while True:
            words = lines[author_index + inc].split()
            # in case loop is finished -> return cif lines.
            # use regular expressions ?
            if len(words) == 0 or words[0] == 'loop_' or words[0][0] == '_':
                return '\n'.join(lines)

            if ((words[0][0] == "'" and words[-1][-1] == "'") or (words[0][0] == '"' and words[-1][-1] == '"')):
                # if quotes are already there, check next line
                inc = inc + 1
            else:
                lines[author_index + inc] = "'" + lines[author_index + inc] + "'"
                inc = inc + 1



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data upf` command."""

import os

import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.commands.cmd_data.cmd_export import data_export, export_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo


@verdi_data.group('upf')
def upf():
    """Manipulate UpfData objects (UPF-format pseudopotentials)."""


@upf.command('uploadfamily')
@click.argument('folder', type=click.Path(exists=True, file_okay=False, resolve_path=True))
@click.argument('group_label', type=click.STRING)
@click.argument('group_description', type=click.STRING)
@click.option(
    '--stop-if-existing',
    is_flag=True,
    default=False,
    help='Interrupt pseudos import if a pseudo was already present in the AiiDA database'
)
@decorators.with_dbenv()
def upf_uploadfamily(folder, group_label, group_description, stop_if_existing):
    """
    Create a new UPF family from a folder of UPF files.

    Returns the numbers of files found and the number of nodes uploaded.

    Call without parameters to get some help.
    """
    from aiida.orm.nodes.data.upf import upload_upf_family
    files_found, files_uploaded = upload_upf_family(folder, group_label, group_description, stop_if_existing)
    echo.echo_success(f'UPF files found: {files_found}. New files uploaded: {files_uploaded}')


@upf.command('listfamilies')
@click.option(
    '-d',
    '--with-description',
    'with_description',
    is_flag=True,
    default=False,
    help='Show also the description for the UPF family'
)
@options.WITH_ELEMENTS()
@decorators.with_dbenv()
def upf_listfamilies(elements, with_description):
    """
    List all UPF families that exist in the database.
    """
    from aiida import orm
    from aiida.plugins import DataFactory

    UpfData = DataFactory('core.upf')  # pylint: disable=invalid-name
    query = orm.QueryBuilder()
    query.append(UpfData, tag='upfdata')
    if elements is not None:
        query.add_filter(UpfData, {'attributes.element': {'in': elements}})
    query.append(orm.UpfFamily, with_node='upfdata', tag='group', project=['label', 'description'])

    query.distinct()
    if query.count() > 0:
        for res in query.dict():
            group_label = res.get('group').get('label')
            group_desc = res.get('group').get('description')
            query = orm.QueryBuilder()
            query.append(orm.Group, tag='thisgroup', filters={'label': {'like': group_label}})
            query.append(UpfData, project=['id'], with_group='thisgroup')

            if with_description:
                description_string = f': {group_desc}'
            else:
                description_string = ''

            echo.echo_success(f'* {group_label} [{query.count()} pseudos]{description_string}')

    else:
        echo.echo_warning('No valid UPF pseudopotential family found.')


@upf.command('exportfamily')
@click.argument('folder', type=click.Path(exists=True, file_okay=False, resolve_path=True))
@arguments.GROUP()
@decorators.with_dbenv()
def upf_exportfamily(folder, group):
    """
    Export a pseudopotential family into a folder.
    Call without parameters to get some help.
    """
    if group.is_empty:
        echo.echo_critical(f'Group<{group.label}> contains no pseudos')

    for node in group.nodes:
        dest_path = os.path.join(folder, node.filename)
        if not os.path.isfile(dest_path):
            with open(dest_path, 'w', encoding='utf8') as handle:
                handle.write(node.get_content())
        else:
            echo.echo_warning(f'File {node.filename} is already present in the destination folder')


@upf.command('import')
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@decorators.with_dbenv()
def upf_import(filename):
    """
    Import a UPF pseudopotential from a file.
    """
    from aiida.orm import UpfData

    node, _ = UpfData.get_or_create(filename)
    echo.echo_success(f'Imported: {node}')


@upf.command('export')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.upf',)))
@options.EXPORT_FORMAT(
    type=click.Choice(['json']),
    default='json',
)
@export_options
@decorators.with_dbenv()
def upf_export(**kwargs):
    """Export `UpfData` object to file."""
    node = kwargs.pop('datum')
    output = kwargs.pop('output')
    fmt = kwargs.pop('fmt')
    force = kwargs.pop('force')

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    data_export(node, output, fmt, other_args=kwargs, overwrite=force)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of caching mechanism and configuration for calculations."""
from collections import namedtuple
from contextlib import contextmanager, suppress
from enum import Enum
import keyword
import re

from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.manage.configuration import get_config_option
from aiida.plugins.entry_point import ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP, ENTRY_POINT_STRING_SEPARATOR

__all__ = ('get_use_cache', 'enable_caching', 'disable_caching')


class ConfigKeys(Enum):
    """Valid keys for caching configuration."""

    DEFAULT = 'caching.default_enabled'
    ENABLED = 'caching.enabled_for'
    DISABLED = 'caching.disabled_for'


class _ContextCache:
    """Cache options, accounting for when in enable_caching or disable_caching contexts."""

    def __init__(self):
        self._default_all = None
        self._enable = []
        self._disable = []

    def clear(self):
        """Clear caching overrides."""
        self.__init__()

    def enable_all(self):
        self._default_all = 'enable'

    def disable_all(self):
        self._default_all = 'disable'

    def enable(self, identifier):
        self._enable.append(identifier)
        with suppress(ValueError):
            self._disable.remove(identifier)

    def disable(self, identifier):
        self._disable.append(identifier)
        with suppress(ValueError):
            self._enable.remove(identifier)

    def get_options(self):
        """Return the options, applying any context overrides."""

        if self._default_all == 'disable':
            return False, [], []

        if self._default_all == 'enable':
            return True, [], []

        default = get_config_option(ConfigKeys.DEFAULT.value)
        enabled = get_config_option(ConfigKeys.ENABLED.value)[:]
        disabled = get_config_option(ConfigKeys.DISABLED.value)[:]

        for ident in self._disable:
            disabled.append(ident)
            with suppress(ValueError):
                enabled.remove(ident)

        for ident in self._enable:
            enabled.append(ident)
            with suppress(ValueError):
                disabled.remove(ident)

        # Check validity of enabled and disabled entries
        try:
            for identifier in enabled + disabled:
                _validate_identifier_pattern(identifier=identifier)
        except ValueError as exc:
            raise exceptions.ConfigurationError('Invalid identifier pattern in enable or disable list.') from exc

        return default, enabled, disabled


_CONTEXT_CACHE = _ContextCache()


@contextmanager
def enable_caching(*, identifier=None):
    """Context manager to enable caching, either for a specific node class, or globally.

    .. warning:: this does not affect the behavior of the daemon, only the local Python interpreter.

    :param identifier: Process type string of the node, or a pattern with '*' wildcard that matches it.
        If not provided, caching is enabled for all classes.
    :type identifier: str
    """
    type_check(identifier, str, allow_none=True)

    if identifier is None:
        _CONTEXT_CACHE.enable_all()
    else:
        _validate_identifier_pattern(identifier=identifier)
        _CONTEXT_CACHE.enable(identifier)
    yield
    _CONTEXT_CACHE.clear()


@contextmanager
def disable_caching(*, identifier=None):
    """Context manager to disable caching, either for a specific node class, or globally.

    .. warning:: this does not affect the behavior of the daemon, only the local Python interpreter.

    :param identifier: Process type string of the node, or a pattern with '*' wildcard that matches it.
        If not provided, caching is disabled for all classes.
    :type identifier: str
    """
    type_check(identifier, str, allow_none=True)

    if identifier is None:
        _CONTEXT_CACHE.disable_all()
    else:
        _validate_identifier_pattern(identifier=identifier)
        _CONTEXT_CACHE.disable(identifier)
    yield
    _CONTEXT_CACHE.clear()


def get_use_cache(*, identifier=None):
    """Return whether the caching mechanism should be used for the given process type according to the configuration.

    :param identifier: Process type string of the node
    :type identifier: str
    :return: boolean, True if caching is enabled, False otherwise
    :raises: `~aiida.common.exceptions.ConfigurationError` if the configuration is invalid, either due to a general
        configuration error, or by defining the class both enabled and disabled
    """
    type_check(identifier, str, allow_none=True)

    default, enabled, disabled = _CONTEXT_CACHE.get_options()

    if identifier is not None:
        type_check(identifier, str)

        enable_matches = [pattern for pattern in enabled if _match_wildcard(string=identifier, pattern=pattern)]
        disable_matches = [pattern for pattern in disabled if _match_wildcard(string=identifier, pattern=pattern)]

        if enable_matches and disable_matches:
            # If both enable and disable have matching identifier, we search for
            # the most specific one. This is determined by checking whether
            # all other patterns match the specific pattern.
            PatternWithResult = namedtuple('PatternWithResult', ['pattern', 'use_cache'])
            most_specific = []
            for specific_pattern in enable_matches:
                if all(
                    _match_wildcard(string=specific_pattern, pattern=other_pattern)
                    for other_pattern in enable_matches + disable_matches
                ):
                    most_specific.append(PatternWithResult(pattern=specific_pattern, use_cache=True))
            for specific_pattern in disable_matches:
                if all(
                    _match_wildcard(string=specific_pattern, pattern=other_pattern)
                    for other_pattern in enable_matches + disable_matches
                ):
                    most_specific.append(PatternWithResult(pattern=specific_pattern, use_cache=False))

            if len(most_specific) > 1:
                raise exceptions.ConfigurationError((
                    'Invalid configuration: multiple matches for identifier {}'
                    ', but the most specific identifier is not unique. Candidates: {}'
                ).format(identifier, [match.pattern for match in most_specific]))
            if not most_specific:
                raise exceptions.ConfigurationError(
                    'Invalid configuration: multiple matches for identifier {}, but none of them is most specific.'.
                    format(identifier)
                )
            return most_specific[0].use_cache
        if enable_matches:
            return True
        if disable_matches:
            return False
    return default


def _match_wildcard(*, string, pattern):
    """
    Helper function to check whether a given name matches a pattern
    which can contain '*' wildcards.
    """
    regexp = '.*'.join(re.escape(part) for part in pattern.split('*'))
    return re.fullmatch(pattern=regexp, string=string) is not None


def _validate_identifier_pattern(*, identifier):
    """
    The identifier (without wildcards) can have one of two forms:

    1.  <group_name><ENTRY_POINT_STRING_SEPARATOR><tail>

        where `group_name` is one of the keys in `ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP`
        and `tail` can be anything _except_ `ENTRY_POINT_STRING_SEPARATOR`.

    2. a fully qualified Python name

        this is a colon-separated string, where each part satisfies
        `part.isidentifier() and not keyword.iskeyword(part)`

    This function checks if an identifier _with_ wildcards can possibly
    match one of these two forms. If it can not, a `ValueError` is raised.

    :param identifier: Process type string, or a pattern with '*' wildcard that matches it.
    :type identifier: str
    """
    common_error_msg = f"Invalid identifier pattern '{identifier}': "
    assert ENTRY_POINT_STRING_SEPARATOR not in '.*'  # The logic of this function depends on this
    # Check if it can be an entry point string
    if identifier.count(ENTRY_POINT_STRING_SEPARATOR) > 1:
        raise ValueError(
            f"{common_error_msg}Can contain at most one entry point string separator '{ENTRY_POINT_STRING_SEPARATOR}'"
        )
    # If there is one separator, it must be an entry point string.
    # Check if the left hand side is a matching pattern
    if ENTRY_POINT_STRING_SEPARATOR in identifier:
        group_pattern, _ = identifier.split(ENTRY_POINT_STRING_SEPARATOR)
        if not any(
            _match_wildcard(string=group_name, pattern=group_pattern)
            for group_name in ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP
        ):
            raise ValueError(
                common_error_msg + "Group name pattern '{}' does not match any of the AiiDA entry point group names.".
                format(group_pattern)
            )
        # The group name pattern matches, and there are no further
        # entry point string separators in the identifier, hence it is
        # a valid pattern.
        return
    # The separator might be swallowed in a wildcard, for example
    # aiida.* or aiida.calculations*
    if '*' in identifier:
        group_part, _ = identifier.split('*', 1)
        if any(group_name.startswith(group_part) for group_name in ENTRY_POINT_GROUP_TO_MODULE_PATH_MAP):
            return
    # Finally, check if it could be a fully qualified Python name
    for identifier_part in identifier.split('.'):
        # If it contains a wildcard, we can not check for keywords.
        # Replacing all wildcards with a single letter must give an
        # identifier - this checks for invalid characters, and that it
        # does not start with a number.
        if '*' in identifier_part:
            if not identifier_part.replace('*', 'a').isidentifier():
                raise ValueError(
                    common_error_msg +
                    f"Identifier part '{identifier_part}' can not match a fully qualified Python name."
                )
        else:
            if not identifier_part.isidentifier():
                raise ValueError(f"{common_error_msg}'{identifier_part}' is not a valid Python identifier.")
            if keyword.iskeyword(identifier_part):
                raise ValueError(f"{common_error_msg}'{identifier_part}' is a reserved Python keyword.")



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Change `db_dbnode.type` for base `Data` types.

The base Data types Bool, Float, Int and Str have been moved in the source code, which means that their
module path changes, which determines the plugin type string which is stored in the databse.
The type string now will have a type string prefix that is unique to each sub type.

Revision ID: django_0009
Revises: django_0008

"""
from alembic import op

revision = 'django_0009'
down_revision = 'django_0008'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.execute(
        """
            UPDATE db_dbnode SET type = 'data.bool.Bool.' WHERE type = 'data.base.Bool.';
            UPDATE db_dbnode SET type = 'data.float.Float.' WHERE type = 'data.base.Float.';
            UPDATE db_dbnode SET type = 'data.int.Int.' WHERE type = 'data.base.Int.';
            UPDATE db_dbnode SET type = 'data.str.Str.' WHERE type = 'data.base.Str.';
            UPDATE db_dbnode SET type = 'data.list.List.' WHERE type = 'data.base.List.';
        """
    )


def downgrade():
    """Migrations for the downgrade."""
    op.execute(
        """
            UPDATE db_dbnode SET type = 'data.base.Bool.' WHERE type = 'data.bool.Bool.';
            UPDATE db_dbnode SET type = 'data.base.Float.' WHERE type = 'data.float.Float.';
            UPDATE db_dbnode SET type = 'data.base.Int.' WHERE type = 'data.int.Int.';
            UPDATE db_dbnode SET type = 'data.base.Str.' WHERE type = 'data.str.Str.';
            UPDATE db_dbnode SET type = 'data.base.List.' WHERE type = 'data.list.List.';
        """
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Top level functions that can be used to launch a Process."""
from typing import Any, Dict, Tuple, Type, Union

from aiida.common import InvalidOperation
from aiida.manage import manager
from aiida.orm import ProcessNode

from .processes.functions import FunctionProcess
from .processes.process import Process, ProcessBuilder
from .utils import instantiate_process, is_process_scoped  # pylint: disable=no-name-in-module

__all__ = ('run', 'run_get_pk', 'run_get_node', 'submit')

TYPE_RUN_PROCESS = Union[Process, Type[Process], ProcessBuilder]  # pylint: disable=invalid-name
# run can also be process function, but it is not clear what type this should be
TYPE_SUBMIT_PROCESS = Union[Process, Type[Process], ProcessBuilder]  # pylint: disable=invalid-name


def run(process: TYPE_RUN_PROCESS, *args: Any, **inputs: Any) -> Dict[str, Any]:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class or process function to run
    :param inputs: the inputs to be passed to the process

    :return: the outputs of the process

    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run(process, *args, **inputs)


def run_get_node(process: TYPE_RUN_PROCESS, *args: Any, **inputs: Any) -> Tuple[Dict[str, Any], ProcessNode]:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class, instance, builder or function to run
    :param inputs: the inputs to be passed to the process

    :return: tuple of the outputs of the process and the process node

    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run_get_node(process, *args, **inputs)


def run_get_pk(process: TYPE_RUN_PROCESS, *args: Any, **inputs: Any) -> Tuple[Dict[str, Any], int]:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class, instance, builder or function to run
    :param inputs: the inputs to be passed to the process

    :return: tuple of the outputs of the process and process node pk

    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run_get_pk(process, *args, **inputs)


def submit(process: TYPE_SUBMIT_PROCESS, **inputs: Any) -> ProcessNode:
    """Submit the process with the supplied inputs to the daemon immediately returning control to the interpreter.

    .. warning: this should not be used within another process. Instead, there one should use the `submit` method of
        the wrapping process itself, i.e. use `self.submit`.

    .. warning: submission of processes requires `store_provenance=True`

    :param process: the process class, instance or builder to submit
    :param inputs: the inputs to be passed to the process

    :return: the calculation node of the process

    """
    # Submitting from within another process requires `self.submit` unless it is a work function, in which case the
    # current process in the scope should be an instance of `FunctionProcess`
    if is_process_scoped() and not isinstance(Process.current(), FunctionProcess):
        raise InvalidOperation('Cannot use top-level `submit` from within another process, use `self.submit` instead')

    runner = manager.get_manager().get_runner()
    assert runner.persister is not None, 'runner does not have a persister'
    assert runner.controller is not None, 'runner does not have a persister'

    process_inited = instantiate_process(runner, process, **inputs)

    # If a dry run is requested, simply forward to `run`, because it is not compatible with `submit`. We choose for this
    # instead of raising, because in this way the user does not have to change the launcher when testing. The same goes
    # for if `remote_folder` is present in the inputs, which means we are importing an already completed calculation.
    if process_inited.metadata.get('dry_run', False) or 'remote_folder' in inputs:
        _, node = run_get_node(process_inited)
        return node

    if not process_inited.metadata.store_provenance:
        raise InvalidOperation('cannot submit a process with `store_provenance=False`')

    runner.persister.save_checkpoint(process_inited)
    process_inited.close()

    # Do not wait for the future's result, because in the case of a single worker this would cock-block itself
    runner.controller.continue_process(process_inited.pid, nowait=False, no_reply=True)

    return process_inited.node


# Allow one to also use run.get_node and run.get_pk as a shortcut, without having to import the functions themselves
run.get_node = run_get_node  # type: ignore[attr-defined]
run.get_pk = run_get_pk  # type: ignore[attr-defined]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for functions to traverse AiiDA graphs."""
import sys
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Mapping, Optional, Set, cast

from numpy import inf

from aiida import orm
from aiida.common import exceptions
from aiida.common.links import GraphTraversalRules, LinkType
from aiida.orm.utils.links import LinkQuadruple
from aiida.tools.graph.age_entities import Basket
from aiida.tools.graph.age_rules import RuleSaveWalkers, RuleSequence, RuleSetWalkers, UpdateRule

if TYPE_CHECKING:
    from aiida.orm.implementation import StorageBackend

if sys.version_info >= (3, 8):
    from typing import TypedDict

    class TraverseGraphOutput(TypedDict, total=False):
        nodes: Set[int]
        links: Optional[Set[LinkQuadruple]]
        rules: Dict[str, bool]
else:
    TraverseGraphOutput = Mapping[str, Any]


def get_nodes_delete(
    starting_pks: Iterable[int],
    get_links: bool = False,
    missing_callback: Optional[Callable[[Iterable[int]], None]] = None,
    backend: Optional['StorageBackend'] = None,
    **traversal_rules: bool
) -> TraverseGraphOutput:
    """
    This function will return the set of all nodes that can be connected
    to a list of initial nodes through any sequence of specified authorized
    links and directions for deletion.

    :param starting_pks: Contains the (valid) pks of the starting nodes.

    :param get_links:
        Pass True to also return the links between all nodes (found + initial).

    :param missing_callback: A callback to handle missing starting_pks or if None raise NotExistent
    	For example to ignore them: ``missing_callback=lambda missing_pks: None``

    :param traversal_rules: graph traversal rules. See :const:`aiida.common.links.GraphTraversalRules` what rule names
        are toggleable and what the defaults are.

    """
    traverse_links = validate_traversal_rules(GraphTraversalRules.DELETE, **traversal_rules)

    traverse_output = traverse_graph(
        starting_pks,
        get_links=get_links,
        backend=backend,
        links_forward=traverse_links['forward'],
        links_backward=traverse_links['backward'],
        missing_callback=missing_callback,
    )

    function_output: TraverseGraphOutput = {
        'nodes': traverse_output['nodes'],
        'links': traverse_output['links'],
        'rules': traverse_links['rules_applied']
    }

    return function_output


def get_nodes_export(
    starting_pks: Iterable[int],
    get_links: bool = False,
    backend: Optional['StorageBackend'] = None,
    **traversal_rules: bool
) -> TraverseGraphOutput:
    """
    This function will return the set of all nodes that can be connected
    to a list of initial nodes through any sequence of specified authorized
    links and directions for export. This will also return the links and
    the traversal rules parsed.

    :param starting_pks: Contains the (valid) pks of the starting nodes.

    :param get_links:
        Pass True to also return the links between all nodes (found + initial).

    :param input_calc_forward: will traverse INPUT_CALC links in the forward direction.
    :param create_backward: will traverse CREATE links in the backward direction.
    :param return_backward: will traverse RETURN links in the backward direction.
    :param input_work_forward: will traverse INPUT_WORK links in the forward direction.
    :param call_calc_backward: will traverse CALL_CALC links in the backward direction.
    :param call_work_backward: will traverse CALL_WORK links in the backward direction.
    """
    traverse_links = validate_traversal_rules(GraphTraversalRules.EXPORT, **traversal_rules)

    traverse_output = traverse_graph(
        starting_pks,
        get_links=get_links,
        backend=backend,
        links_forward=traverse_links['forward'],
        links_backward=traverse_links['backward']
    )

    function_output: TraverseGraphOutput = {
        'nodes': traverse_output['nodes'],
        'links': traverse_output['links'],
        'rules': traverse_links['rules_applied']
    }

    return function_output


def validate_traversal_rules(
    ruleset: GraphTraversalRules = GraphTraversalRules.DEFAULT, **traversal_rules: bool
) -> dict:
    """
    Validates the keywords with a ruleset template and returns a parsed dictionary
    ready to be used.

    :param ruleset: Ruleset template used to validate the set of rules.
    :param input_calc_forward: will traverse INPUT_CALC links in the forward direction.
    :param input_calc_backward: will traverse INPUT_CALC links in the backward direction.
    :param create_forward: will traverse CREATE links in the forward direction.
    :param create_backward: will traverse CREATE links in the backward direction.
    :param return_forward: will traverse RETURN links in the forward direction.
    :param return_backward: will traverse RETURN links in the backward direction.
    :param input_work_forward: will traverse INPUT_WORK links in the forward direction.
    :param input_work_backward: will traverse INPUT_WORK links in the backward direction.
    :param call_calc_forward: will traverse CALL_CALC links in the forward direction.
    :param call_calc_backward: will traverse CALL_CALC links in the backward direction.
    :param call_work_forward: will traverse CALL_WORK links in the forward direction.
    :param call_work_backward: will traverse CALL_WORK links in the backward direction.
    """
    if not isinstance(ruleset, GraphTraversalRules):
        raise TypeError(
            f'ruleset input must be of type aiida.common.links.GraphTraversalRules\ninstead, it is: {type(ruleset)}'
        )

    rules_applied: Dict[str, bool] = {}
    links_forward: List[LinkType] = []
    links_backward: List[LinkType] = []

    for name, rule in ruleset.value.items():

        follow = rule.default

        if name in traversal_rules:

            if not rule.toggleable:
                raise ValueError(f'input rule {name} is not toggleable for ruleset {ruleset}')

            follow = traversal_rules.pop(name)

            if not isinstance(follow, bool):
                raise ValueError(f'the value of rule {name} must be boolean, but it is: {follow}')

        if follow:

            if rule.direction == 'forward':
                links_forward.append(rule.link_type)
            elif rule.direction == 'backward':
                links_backward.append(rule.link_type)
            else:
                raise exceptions.InternalError(f'unrecognized direction `{rule.direction}` for graph traversal rule')

        rules_applied[name] = follow

    if traversal_rules:
        error_message = f"unrecognized keywords: {', '.join(traversal_rules.keys())}"
        raise exceptions.ValidationError(error_message)

    valid_output = {
        'rules_applied': rules_applied,
        'forward': links_forward,
        'backward': links_backward,
    }

    return valid_output


def traverse_graph(
    starting_pks: Iterable[int],
    max_iterations: Optional[int] = None,
    get_links: bool = False,
    links_forward: Iterable[LinkType] = (),
    links_backward: Iterable[LinkType] = (),
    missing_callback: Optional[Callable[[Iterable[int]], None]] = None,
    backend: Optional['StorageBackend'] = None
) -> TraverseGraphOutput:
    """
    This function will return the set of all nodes that can be connected
    to a list of initial nodes through any sequence of specified links.
    Optionally, it may also return the links that connect these nodes.

    :param starting_pks: Contains the (valid) pks of the starting nodes.

    :param max_iterations:
        The number of iterations to apply the set of rules (a value of 'None' will
        iterate until no new nodes are added).

    :param get_links: Pass True to also return the links between all nodes (found + initial).

    :param links_forward: List with all the links that should be traversed in the forward direction.
    :param links_backward: List with all the links that should be traversed in the backward direction.

    :param missing_callback: A callback to handle missing starting_pks or if None raise NotExistent
    """
    # pylint: disable=too-many-locals,too-many-statements,too-many-branches

    if max_iterations is None:
        max_iterations = cast(int, inf)
    elif not (isinstance(max_iterations, int) or max_iterations is inf):
        raise TypeError('Max_iterations has to be an integer or infinity')

    linktype_list = []
    for linktype in links_forward:
        if not isinstance(linktype, LinkType):
            raise TypeError(f'links_forward should contain links, but one of them is: {type(linktype)}')
        linktype_list.append(linktype.value)
    filters_forwards = {'type': {'in': linktype_list}}

    linktype_list = []
    for linktype in links_backward:
        if not isinstance(linktype, LinkType):
            raise TypeError(f'links_backward should contain links, but one of them is: {type(linktype)}')
        linktype_list.append(linktype.value)
    filters_backwards = {'type': {'in': linktype_list}}

    if not isinstance(starting_pks, Iterable):  # pylint: disable=isinstance-second-argument-not-valid-type
        raise TypeError(f'starting_pks must be an iterable\ninstead, it is {type(starting_pks)}')

    if any(not isinstance(pk, int) for pk in starting_pks):
        raise TypeError(f'one of the starting_pks is not of type int:\n {starting_pks}')
    operational_set = set(starting_pks)

    if not operational_set:
        if get_links:
            return {'nodes': set(), 'links': set()}
        return {'nodes': set(), 'links': None}

    query_nodes = orm.QueryBuilder(backend=backend)
    query_nodes.append(orm.Node, project=['id'], filters={'id': {'in': operational_set}})
    existing_pks = set(query_nodes.all(flat=True))
    missing_pks = operational_set.difference(existing_pks)
    if missing_pks and missing_callback is None:
        raise exceptions.NotExistent(
            f'The following pks are not in the database and must be pruned before this call: {missing_pks}'
        )
    elif missing_pks and missing_callback is not None:
        missing_callback(missing_pks)

    rules = []
    basket = Basket(nodes=existing_pks)

    # When max_iterations is finite, the order of traversal may affect the result
    # (its not the same to first go backwards and then forwards than vice-versa)
    # In order to make it order-independent, the result of the first operation needs
    # to be stashed and the second operation must be performed only on the nodes
    # that were already in the set at the begining of the iteration: this way, both
    # rules are applied on the same set of nodes and the order doesn't matter.
    # The way to do this is saving and seting the walkers at the right moments only
    # when both forwards and backwards rules are present.
    if links_forward and links_backward:
        stash = basket.get_template()
        rules += [RuleSaveWalkers(stash)]

    if links_forward:
        query_outgoing = orm.QueryBuilder(backend=backend)
        query_outgoing.append(orm.Node, tag='sources')
        query_outgoing.append(orm.Node, edge_filters=filters_forwards, with_incoming='sources')
        rule_outgoing = UpdateRule(query_outgoing, max_iterations=1, track_edges=get_links)
        rules += [rule_outgoing]

    if links_forward and links_backward:
        rules += [RuleSetWalkers(stash)]

    if links_backward:
        query_incoming = orm.QueryBuilder(backend=backend)
        query_incoming.append(orm.Node, tag='sources')
        query_incoming.append(orm.Node, edge_filters=filters_backwards, with_outgoing='sources')
        rule_incoming = UpdateRule(query_incoming, max_iterations=1, track_edges=get_links)
        rules += [rule_incoming]

    rulesequence = RuleSequence(rules, max_iterations=max_iterations)

    results = rulesequence.run(basket)

    output: TraverseGraphOutput = {}
    output['nodes'] = results.nodes.keyset
    output['links'] = None
    if get_links:
        output['links'] = results['nodes_nodes'].keyset

    return output



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This is the sqlite DB schema, coresponding to the `main_0000` revision of the `sqlite_zip` backend,
see: `versions/main_0000_initial.py`

For normal operation of the archive,
we auto-generate the schema from the models in ``aiida.storage.psql_dos.models``.
However, when migrating an archive from the old format, we require a fixed revision of the schema.

The only difference between the PostGreSQL schema and SQLite one,
is the replacement of ``JSONB`` with ``JSON``, and ``UUID`` with ``CHAR(32)``.
"""
from sqlalchemy import ForeignKey, MetaData, orm
from sqlalchemy.dialects.sqlite import JSON
from sqlalchemy.schema import Column, UniqueConstraint
from sqlalchemy.types import CHAR, Boolean, DateTime, Integer, String, Text

from aiida.common import timezone
from aiida.common.utils import get_new_uuid

# see https://alembic.sqlalchemy.org/en/latest/naming.html
naming_convention = (
    ('pk', '%(table_name)s_pkey'),
    ('ix', 'ix_%(table_name)s_%(column_0_N_label)s'),
    ('uq', 'uq_%(table_name)s_%(column_0_N_name)s'),
    ('ck', 'ck_%(table_name)s_%(constraint_name)s'),
    ('fk', 'fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s'),
)

ArchiveV1Base = orm.declarative_base(metadata=MetaData(naming_convention=dict(naming_convention)))


class DbAuthInfo(ArchiveV1Base):
    """Class that keeps the authentication data."""

    __tablename__ = 'db_dbauthinfo'
    __table_args__ = (UniqueConstraint('aiidauser_id', 'dbcomputer_id'),)

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    aiidauser_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True
    )
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True
    )
    _metadata = Column('metadata', JSON, default=dict, nullable=True)
    auth_params = Column(JSON, default=dict, nullable=True)
    enabled = Column(Boolean, default=True, nullable=True)


class DbComment(ArchiveV1Base):
    """Class to store comments."""

    __tablename__ = 'db_dbcomment'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    dbnode_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True
    )
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    mtime = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True
    )
    content = Column(Text, default='', nullable=True)


class DbComputer(ArchiveV1Base):
    """Class to store computers."""
    __tablename__ = 'db_dbcomputer'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    label = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255), default='', nullable=True)
    description = Column(Text, default='', nullable=True)
    scheduler_type = Column(String(255), default='', nullable=True)
    transport_type = Column(String(255), default='', nullable=True)
    _metadata = Column('metadata', JSON, default=dict, nullable=True)


class DbGroupNodes(ArchiveV1Base):
    """Class to store join table for group -> nodes."""

    __tablename__ = 'db_dbgroup_dbnodes'
    __table_args__ = (UniqueConstraint('dbgroup_id', 'dbnode_id'),)

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    dbnode_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )
    dbgroup_id = Column(
        Integer, ForeignKey('db_dbgroup.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )


class DbGroup(ArchiveV1Base):
    """Class to store groups."""

    __tablename__ = 'db_dbgroup'
    __table_args__ = (UniqueConstraint('label', 'type_string'),)

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    label = Column(String(255), nullable=False, index=True)
    type_string = Column(String(255), default='', nullable=True, index=True)
    time = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    description = Column(Text, default='', nullable=True)
    extras = Column(JSON, default=dict, nullable=False)
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True
    )


class DbLog(ArchiveV1Base):
    """Class to store logs."""

    __tablename__ = 'db_dblog'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    time = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    loggername = Column(String(255), default='', nullable=True, index=True)
    levelname = Column(String(50), default='', nullable=True, index=True)
    dbnode_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED', ondelete='CASCADE'),
        nullable=False,
        index=True
    )
    message = Column(Text(), default='', nullable=True)
    _metadata = Column('metadata', JSON, default=dict, nullable=True)


class DbNode(ArchiveV1Base):
    """Class to store nodes."""

    __tablename__ = 'db_dbnode'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    node_type = Column(String(255), default='', nullable=False, index=True)
    process_type = Column(String(255), index=True)
    label = Column(String(255), default='', index=True, nullable=True)
    description = Column(Text(), default='', nullable=True)
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=True, index=True)
    mtime = Column(DateTime(timezone=True), default=timezone.now, nullable=True, index=True)
    attributes = Column(JSON)
    extras = Column(JSON)
    repository_metadata = Column(JSON, nullable=False, default=dict, server_default='{}')
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially='DEFERRED', ondelete='RESTRICT'),
        nullable=True,
        index=True
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially='DEFERRED', ondelete='restrict'),
        nullable=False,
        index=True
    )


class DbLink(ArchiveV1Base):
    """Class to store links between nodes."""

    __tablename__ = 'db_dblink'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    input_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )
    output_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True
    )
    label = Column(String(255), default='', nullable=False, index=True)
    type = Column(String(255), nullable=False, index=True)


class DbUser(ArchiveV1Base):
    """Class to store users."""

    __tablename__ = 'db_dbuser'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    email = Column(String(254), nullable=False, unique=True)
    first_name = Column(String(254), default='', nullable=True)
    last_name = Column(String(254), default='', nullable=True)
    institution = Column(String(254), default='', nullable=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=global-statement,unused-argument
"""Provide a singleton progress reporter implementation.

The interface is inspired by `tqdm <https://github.com/tqdm/tqdm>`,
and indeed a valid implementation is::

    from tqdm import tqdm
    set_progress_reporter(tqdm, bar_format='{l_bar}{bar}{r_bar}')

"""
from functools import partial
from types import TracebackType
from typing import Any, Callable, Optional, Type

__all__ = (
    'get_progress_reporter', 'set_progress_reporter', 'set_progress_bar_tqdm', 'ProgressReporterAbstract',
    'TQDM_BAR_FORMAT', 'create_callback'
)

TQDM_BAR_FORMAT = '{desc:40.40}{percentage:6.1f}%|{bar}| {n_fmt}/{total_fmt}'


class ProgressReporterAbstract:
    """An abstract class for incrementing a progress reporter.

    This class provides the base interface for any `ProgressReporter` class.

    Example Usage::

        with ProgressReporter(total=10, desc="A process:") as progress:
            for i in range(10):
                progress.set_description_str(f"A process: {i}")
                progress.update()

    """

    def __init__(self, *, total: int, desc: Optional[str] = None, **kwargs: Any):
        """Initialise the progress reporting contextmanager.

        :param total: The number of expected iterations.
        :param desc: A description of the process

        """
        self._total = total
        self._desc = desc
        self._increment: int = 0

    @property
    def total(self) -> int:
        """Return the total iterations expected."""
        return self._total

    @property
    def desc(self) -> Optional[str]:
        """Return the description of the process."""
        return self._desc

    @property
    def n(self) -> int:  # pylint: disable=invalid-name
        """Return the current iteration."""
        # note using `n` as the attribute name is necessary for compatibility with tqdm
        return self._increment

    def __enter__(self) -> 'ProgressReporterAbstract':
        """Enter the contextmanager."""
        return self

    def __exit__(
        self, exctype: Optional[Type[BaseException]], excinst: Optional[BaseException], exctb: Optional[TracebackType]
    ):
        """Exit the contextmanager."""
        return False

    def set_description_str(self, text: Optional[str] = None, refresh: bool = True):
        """Set the text shown by the progress reporter.

        :param text: The text to show
        :param refresh: Force refresh of the progress reporter

        """
        self._desc = text

    def update(self, n: int = 1):  # pylint: disable=invalid-name
        """Update the progress counter.

        :param n: Increment to add to the internal counter of iterations

        """
        self._increment += n

    def reset(self, total: Optional[int] = None):
        """Resets current iterations to 0.

        :param total: If not None, update number of expected iterations.

        """
        self._increment = 0
        if total is not None:
            self._total = total


class ProgressReporterNull(ProgressReporterAbstract):
    """A null implementation of the progress reporter.

    This implementation does not output anything.
    """


PROGRESS_REPORTER: Type[ProgressReporterAbstract] = ProgressReporterNull  # pylint: disable=invalid-name


def get_progress_reporter() -> Type[ProgressReporterAbstract]:
    """Return the progress reporter

    Example Usage::

        with get_progress_reporter()(total=10, desc="A process:") as progress:
            for i in range(10):
                progress.set_description_str(f"A process: {i}")
                progress.update()

    """
    global PROGRESS_REPORTER  # pylint: disable=global-variable-not-assigned
    return PROGRESS_REPORTER


def set_progress_reporter(reporter: Optional[Type[ProgressReporterAbstract]] = None, **kwargs: Any):
    """Set the progress reporter implementation

    :param reporter: A progress reporter for a process.  If None, reset to ``ProgressReporterNull``.

    :param kwargs: If present, set a partial function with these kwargs

    The reporter should be a context manager that implements the
    :func:`~aiida.common.progress_reporter.ProgressReporterAbstract` interface.

    Example Usage::

        set_progress_reporter(ProgressReporterNull)
        with get_progress_reporter()(total=10, desc="A process:") as progress:
            for i in range(10):
                progress.set_description_str(f"A process: {i}")
                progress.update()

    """
    global PROGRESS_REPORTER
    if reporter is None:
        PROGRESS_REPORTER = ProgressReporterNull
    elif kwargs:
        PROGRESS_REPORTER = partial(reporter, **kwargs)  # type: ignore
    else:
        PROGRESS_REPORTER = reporter


def set_progress_bar_tqdm(bar_format: Optional[str] = TQDM_BAR_FORMAT, leave: Optional[bool] = False, **kwargs: Any):
    """Set a `tqdm <https://github.com/tqdm/tqdm>`__ implementation of the progress reporter interface.

    See :func:`~aiida.common.progress_reporter.set_progress_reporter` for details.

    :param bar_format: Specify a custom bar string format.
    :param leave: If True, keeps all traces of the progressbar upon termination of iteration.
            If `None`, will leave only if `position` is `0`.
    :param kwargs: pass to the tqdm init

    """
    from tqdm import tqdm
    set_progress_reporter(tqdm, bar_format=bar_format, leave=leave, **kwargs)


def create_callback(progress_reporter: ProgressReporterAbstract) -> Callable[[str, Any], None]:
    """Create a callback function to update the progress reporter.

    :returns: a callback to report on the process, ``callback(action, value)``,
        with the following callback signatures:

        - ``callback('init', {'total': <int>, 'description': <str>})``,
            to reset the progress with a new total iterations and description
        - ``callback('update', <int>)``,
            to update the progress by a certain number of iterations

    """

    def _callback(action: str, value: Any):
        if action == 'init':
            progress_reporter.reset(value['total'])
            progress_reporter.set_description_str(value['description'])
        elif action == 'update':
            progress_reporter.update(value)

    return _callback



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Alter columns to be non-nullable (to bring inline with psql_dos main_0001).

Revision ID: main_0000b
Revises: main_0000a
Create Date: 2022-03-04

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'main_0000b'
down_revision = 'main_0000a'
branch_labels = None
depends_on = None


def upgrade():
    """Upgrade database schema."""
    # see https://alembic.sqlalchemy.org/en/latest/batch.html#running-batch-migrations-for-sqlite-and-other-databases
    # for why we run these in batches
    with op.batch_alter_table('db_dbauthinfo') as batch_op:
        batch_op.alter_column('aiidauser_id', existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column('dbcomputer_id', existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column('metadata', existing_type=sa.JSON(), nullable=False)
        batch_op.alter_column('auth_params', existing_type=sa.JSON(), nullable=False)
        batch_op.alter_column('enabled', existing_type=sa.BOOLEAN(), nullable=False)

    with op.batch_alter_table('db_dbcomment') as batch_op:
        batch_op.alter_column('dbnode_id', existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column('user_id', existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column('content', existing_type=sa.TEXT(), nullable=False)
        batch_op.alter_column('ctime', existing_type=sa.DateTime(timezone=True), nullable=False)
        batch_op.alter_column('mtime', existing_type=sa.DateTime(timezone=True), nullable=False)

    with op.batch_alter_table('db_dbcomputer') as batch_op:
        batch_op.alter_column('description', existing_type=sa.TEXT(), nullable=False)
        batch_op.alter_column('hostname', existing_type=sa.String(255), nullable=False)
        batch_op.alter_column('metadata', existing_type=sa.JSON(), nullable=False)
        batch_op.alter_column('scheduler_type', existing_type=sa.String(255), nullable=False)
        batch_op.alter_column('transport_type', existing_type=sa.String(255), nullable=False)

    with op.batch_alter_table('db_dbgroup') as batch_op:
        batch_op.alter_column('description', existing_type=sa.TEXT(), nullable=False)
        batch_op.alter_column('time', existing_type=sa.DateTime(timezone=True), nullable=False)
        batch_op.alter_column('type_string', existing_type=sa.String(255), nullable=False)

    with op.batch_alter_table('db_dblog') as batch_op:
        batch_op.alter_column('levelname', existing_type=sa.String(50), nullable=False)
        batch_op.alter_column('loggername', existing_type=sa.String(255), nullable=False)
        batch_op.alter_column('message', existing_type=sa.TEXT(), nullable=False)
        batch_op.alter_column('time', existing_type=sa.DateTime(timezone=True), nullable=False)
        batch_op.alter_column('metadata', existing_type=sa.JSON(), nullable=False)

    with op.batch_alter_table('db_dbnode') as batch_op:
        batch_op.alter_column('ctime', existing_type=sa.DateTime(timezone=True), nullable=False)
        batch_op.alter_column('description', existing_type=sa.TEXT(), nullable=False)
        batch_op.alter_column('label', existing_type=sa.String(255), nullable=False)
        batch_op.alter_column('mtime', existing_type=sa.DateTime(timezone=True), nullable=False)

    with op.batch_alter_table('db_dbuser') as batch_op:
        batch_op.alter_column('first_name', existing_type=sa.String(254), nullable=False)
        batch_op.alter_column('last_name', existing_type=sa.String(254), nullable=False)
        batch_op.alter_column('institution', existing_type=sa.String(254), nullable=False)


def downgrade():
    """Downgrade database schema."""
    raise NotImplementedError('Downgrade of main_0000b.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Various decorators useful for creating verdi commands, for example loading the dbenv lazily.

Always avoids trying to load the dbenv twice. When it has to be loaded,
a spinner ASCII widget is displayed.

Provides:
 * with_dbenv, a decorator to frontload the dbenv for functions
 * dbenv, a contextmanager to load the dbenv only if a specific
    code branch gets visited and possibly avoiding the overhead if not

"""
from contextlib import contextmanager

from click_spinner import spinner
from wrapt import decorator

from . import echo

DAEMON_NOT_RUNNING_DEFAULT_MESSAGE = 'daemon is not running'

__all__ = ('with_dbenv', 'dbenv', 'only_if_daemon_running')


def load_backend_if_not_loaded():
    """Load the database backend environment for the currently loaded profile.

    If no profile has been loaded yet, the default profile will be loaded first. A spinner will be shown during both
    actions to indicate that the function is working and has not crashed, since loading can take a second.
    """
    from aiida.manage import get_manager

    manager = get_manager()

    if manager.get_profile() is None or not manager.profile_storage_loaded:
        with spinner():
            manager.load_profile()  # This will load the default profile if no profile has already been loaded
            manager.get_profile_storage()  # This will load the backend of the loaded profile, if not already loaded


def with_dbenv():
    """Function decorator that will load the database environment for the currently loaded profile.

    .. note:: if no profile has been loaded yet, the default profile will be loaded first.

    Example::

        @with_dbenv()
        def create_node():
            from aiida.orm import Int  # note the local import
            node = Int(1).store()
    """

    @decorator
    def wrapper(wrapped, _, args, kwargs):
        """The wrapped function that should be called after having loaded the backend."""
        from aiida.common.exceptions import ConfigurationError, IntegrityError

        try:
            load_backend_if_not_loaded()
        except (IntegrityError, ConfigurationError) as exception:
            echo.echo_critical(str(exception))

        return wrapped(*args, **kwargs)

    return wrapper


@contextmanager
def dbenv():
    """
    Loads the dbenv for a specific region of code, does not unload afterwards

    Only use when it makes it possible to avoid loading the dbenv for certain
    code paths

    Good Example::

        # do this
        @click.command()
        @click.option('--with-db', is_flag=True)
        def profile_info(with_db):
            # read the config file
            click.echo(profile_config)

            # load the db only if necessary
            if with_db:
                with dbenv():
                    # gather db statistics for the profile
                    click.echo(db_statistics)

    This will run very fast without the --with-db flag and slow only if database info is requested

    Do not use if you will end up loading the dbenv anyway

    Bad Example::

        # don't do this
        def my_function():
            with dbenv():
                # read from db

            # do db unrelated stuff
    """
    load_backend_if_not_loaded()
    yield


def only_if_daemon_running(echo_function=echo.echo_critical, message=None):
    """Function decorator for CLI command to print critical error and exit automatically when daemon is not running.

    The error printing and exit behavior can be controlled with the decorator keyword arguments. The default message
    that is printed can be overridden as well as the echo function that is to be used. By default it uses the
    `aiida.cmdline.utils.echo.echo_critical` function which automatically aborts the command. The function can be
    substituted by for example `aiida.cmdline.utils.echo.echo_warning` to instead print just a warning and continue.

    Example::

        @only_if_daemon_running(echo_function=echo.echo_warning, message='beware that the daemon is not running')
        def create_node():
            pass

    :param echo_function: echo function to issue the message, should be from `aiida.cmdline.utils.echo`
    :param message: optional message to override the default message
    """
    if message is None:
        message = DAEMON_NOT_RUNNING_DEFAULT_MESSAGE

    @decorator
    def wrapper(wrapped, _, args, kwargs):
        """If daemon pid file is not found / empty, echo message and call decorated function."""
        from aiida.engine.daemon.client import get_daemon_client

        daemon_client = get_daemon_client()

        if not daemon_client.get_daemon_pid():
            echo_function(message)

        return wrapped(*args, **kwargs)

    return wrapper


@decorator
def check_circus_zmq_version(wrapped, _, args, kwargs):
    """
    Function decorator to check for the right ZMQ version before trying to run circus.

    Example::

        @click.command()
        @check_circus_zmq_version
        def do_circus_stuff():
            pass
    """
    import zmq
    try:
        zmq_version = [int(part) for part in zmq.__version__.split('.')[:2]]
        if len(zmq_version) < 2:
            raise ValueError()
    except (AttributeError, ValueError):
        echo.echo_critical('Unknown PyZQM version - aborting...')

    if zmq_version[0] < 13 or (zmq_version[0] == 13 and zmq_version[1] < 1):
        echo.echo_critical('AiiDA daemon needs PyZMQ >= 13.1.0 to run - aborting...')

    return wrapped(*args, **kwargs)


def deprecated_command(message):
    """Function decorator that will mark a click command as deprecated when invoked.

        Example::

            @click.command()
            @deprecated_command('This command has been deprecated in AiiDA v1.0, please use 'foo' instead.)
            def mycommand():
                pass
    """

    @decorator
    def wrapper(wrapped, _, args, kwargs):
        """Echo a deprecation warning before doing anything else."""
        from textwrap import wrap

        from aiida.cmdline.utils import templates

        template = templates.env.get_template('deprecated.tpl')
        width = 80
        echo.echo(template.render(msg=wrap(message, width - 4), width=width))

        return wrapped(*args, **kwargs)

    return wrapper
